# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from pluginsmanager.model.midi_port import MidiPort


class MidiOutput(MidiPort, metaclass=ABCMeta):
    """
    MidiOutput is the medium in which the midi output processed
    by the effect is returned.

    For obtains the outputs::

        >>> cctonode
        <Lv2Effect object as 'CC2Note'  active at 0x7efe5480af28>
        >>> cctonode.outputs
        (<Lv2MidiOutput object as MIDI Out at 0x7efe5420eeb8>,)

        >>> midi_output = cctonode.midi_outputs[0]
        >>> midi_output
        <Lv2Output object as Out L at 0x7fd58c58a438>

        >>> symbol = midi_output.symbol
        >>> symbol
        'midiout'

        >>> cctonode.midi_outputs[symbol] == midi_output
        True

    For connections between effects, see :meth:`~pluginsmanager.model.pedalboard.Pedalboard.connect()`
    and :meth:`~pluginsmanager.model.pedalboard.Pedalboard.disconnect()` :class:`.Pedalboard` class methods.

    :param Effect effect: Effect that contains the output
    """

    @property
    def index(self):
        """
        :return: Output index in the your effect
        """
        return self.effect.midi_outputs.index(self)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.bank import Bank
from pluginsmanager.model.lv2.lv2_effect_builder import Lv2EffectBuilder as Lv2LilvEffectBuilder
from pluginsmanager.model.pedalboard import Pedalboard
from pluginsmanager.util.builder.lv2_json_builder import Lv2AudioPortBuilder, Lv2EffectBuilder
from pluginsmanager.util.builder.system_json_builder import SystemAudioPortBuilder


class PersistenceDecoderError(Exception):
    pass


class PersistenceDecoder(object):

    def __init__(self, system_effect):
        self.system_effect = system_effect

    def read(self, json):
        reader = BankReader(self.system_effect)

        return reader.read(json)


class Reader(object):
    def __init__(self, system_effect):
        self.system_effect = system_effect

    def read(self, json):
        pass


class BankReader(Reader):

    def read(self, json):
        bank = Bank(json['name'])

        pedalboard_reader = PedalboardReader(self.system_effect)
        for pedalboard_json in json['pedalboards']:
            bank.append(pedalboard_reader.read(pedalboard_json))

        return bank


class PedalboardReader(Reader):

    def read(self, json):
        pedalboard = Pedalboard(json['name'])

        effect_reader = EffectReader(self.system_effect)
        for effect_json in json['effects']:
            pedalboard.append(effect_reader.read(effect_json))

        connection_reader = ConnectionReader(pedalboard, self.system_effect)
        for connection_json in json['connections']:
            port_output, port_input = connection_reader.read(connection_json)
            pedalboard.connect(port_output, port_input)

        if 'data' in json:
            pedalboard.data = json['data']

        return pedalboard


class EffectReader(Reader):

    def __init__(self, system_effect):
        super(EffectReader, self).__init__(system_effect)
        self.builder = Lv2LilvEffectBuilder()

    def read(self, json):
        return self.generate_builder(json).build(json)

    def generate_builder(self, json):
        technology = json['technology']

        if technology == 'lv2':
            return Lv2EffectBuilder(self.builder)
        else:
            raise PersistenceDecoderError('Unknown effect technology: ' + technology)


class ConnectionReader(Reader):

    def __init__(self, pedalboard, system_effect):
        super(ConnectionReader, self).__init__(system_effect)
        self.pedalboard = pedalboard

    def read(self, json):
        if json['type'] == 'audio':
            connection_output = self.generate_builder(json, 'output').build_output(json['output'])
            connection_input = self.generate_builder(json, 'input').build_input(json['input'])
        else:
            connection_output = self.generate_builder(json, 'output').build_midi_output(json['output'])
            connection_input = self.generate_builder(json, 'input').build_midi_input(json['input'])

        return connection_output, connection_input

    def generate_builder(self, json, audio_port):
        """
        :return AudioPortBuilder
        """
        if 'effect' in json[audio_port]:
            return Lv2AudioPortBuilder(self.pedalboard)
        else:
            return SystemAudioPortBuilder(self.system_effect)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DictTuple(tuple):
    """
    Dict tuple is a union with dicts and tuples. It's possible obtains an element
    by index or by a key

    The key is not been a int or long instance

    Based in http://jfine-python-classes.readthedocs.io/en/latest/subclass-tuple.html

    :param iterable elements: Elements for the tuple
    :param lambda key_function: Function mapper: it obtains an element and
                                returns your key.
    """

    def __new__(cls, elements, key_function):
        return tuple.__new__(DictTuple, tuple(elements))

    def __init__(self, elements, key_function):
        self._dict = dict(
            (key_function(element), element) for element in elements
        )

    def __getitem__(self, index):
        if isinstance(index, int):
            return super(DictTuple, self).__getitem__(index)

        else:
            return self._dict[index]



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.output import Output
from pluginsmanager.model.system.system_port_mixing import SystemPortMixing


class SystemOutput(SystemPortMixing, Output):
    """
    Representation of a System output audio port instance.

    For general input use, see :class:`.Output` class documentation.

    :param SystemEffect effect: Effect that contains the input
    :param string symbol: *output audio port* symbol identifier
    """

    def __init__(self, effect, symbol):
        super(SystemOutput, self).__init__(effect)
        self._symbol = symbol

    @property
    def symbol(self):
        return self._symbol



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.observer.scope import ManagerScopes
from pluginsmanager.observer.updates_observer import UpdatesObserver


class ObserverManager(UpdatesObserver):
    def __init__(self):
        super(ObserverManager, self).__init__()
        self.observers = []
        self._scope = ManagerScopes()

    def enter_scope(self, observer):
        """
        Open a observer scope.

        Informs that changes occurs by the ``observer`` and isn't necessary
        informs the changes for observer

        :param UpdatesObserver observer: Observer that causes changes
        """
        self._scope.enter(observer)

    def exit_scope(self):
        """
        Closes the last observer scope added
        """
        self._scope.exit()

    @property
    def scope(self):
        return self._scope.current.identifier

    def append(self, observer):
        self.observers.append(observer)

    def on_bank_updated(self, bank, update_type, index, origin, **kwargs):
        for observer in self.observers:
            if observer != self.scope:
                observer.on_bank_updated(bank, update_type, index=index, origin=origin, **kwargs)

    def on_pedalboard_updated(self, pedalboard, update_type, index, origin, **kwargs):
        for observer in self.observers:
            if observer != self.scope:
                observer.on_pedalboard_updated(pedalboard, update_type, index=index, origin=origin, **kwargs)

    def on_effect_updated(self, effect, update_type, index, origin, **kwargs):
        for observer in self.observers:
            if observer != self.scope:
                observer.on_effect_updated(effect, update_type, index=index, origin=origin, **kwargs)

    def on_effect_status_toggled(self, effect, **kwargs):
        for observer in self.observers:
            if observer != self.scope:
                observer.on_effect_status_toggled(effect, **kwargs)

    def on_param_value_changed(self, param, **kwargs):
        for observer in self.observers:
            if observer != self.scope:
                observer.on_param_value_changed(param, **kwargs)

    def on_connection_updated(self, connection, update_type, pedalboard, **kwargs):
        for observer in self.observers:
            if observer != self.scope:
                observer.on_connection_updated(connection, update_type, pedalboard=pedalboard, **kwargs)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pluginsmanager.observer.mod_host.connection import Connection
from pluginsmanager.observer.mod_host.protocol_parser import ProtocolParser


class Host:
    """
    Bridge between *mod-host* API and *mod-host* process
    """

    def __init__(self, address='localhost', port=5555):
        self.connection = None
        self.connection_fd = None

        # mod-host works if only exists two connections:
        #  - For communication
        try:
            self.connection = Connection(port, address)
        except ConnectionRefusedError as e:
            raise ConnectionRefusedError(str(e) + '. Do you starts mod-host?') from e

        #  - For callback?
        try:
            self.connection_fd = Connection(port+1, address)
        except ConnectionRefusedError as e:
            logging.info('Mod-host - Feedback socket is not enabled')
            logging.info('           Try start Mod-host using: mod-host -f {}'.format(port+1))

        self.instance_index = 0

    def add(self, effect):
        """
        Add an LV2 plugin encapsulated as a jack client

        :param Lv2Effect effect: Effect that will be loaded as LV2 plugin encapsulated
        """
        effect.instance = self.instance_index
        self.instance_index += 1

        self.connection.send(ProtocolParser.add(effect))

    def remove(self, effect):
        """
        Remove an LV2 plugin instance (and also the jack client)

        :param Lv2Effect effect: Effect that your jack client encapsulated will removed
        """
        self.connection.send(ProtocolParser.remove(effect))

    def connect(self, connection):
        """
        Connect two effect audio ports

        :param pluginsmanager.model.connection.Connection connection: Connection with the two effect audio ports (output and input)
        """
        self.connection.send(ProtocolParser.connect(connection))

    def disconnect(self, connection):
        """
        Disconnect two effect audio ports

        :param pluginsmanager.model.connection.Connection connection: Connection with the two effect audio ports (output and input)
        """
        self.connection.send(ProtocolParser.disconnect(connection))

    def set_param_value(self, param):
        """
        Set a value to given control

        :param Lv2Param param: Param that the value will be updated
        """
        self.connection.send(ProtocolParser.param_set(param))

    def set_status(self, effect):
        """
        Toggle effect processing

        :param Lv2Effect effect: Effect with the status updated
        """
        self.connection.send(ProtocolParser.bypass(effect))

    def quit(self):
        """
        Quit the connection with mod-host and
        stop the mod-host process
        """
        self.connection.send(ProtocolParser.quit())
        self.close()

    def close(self):
        """
        Quit the connection with mod-host
        """
        if self.connection is not None:
            self.connection.close()
        if self.connection_fd is not None:
            self.connection_fd.close()



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.observer.observable_list import ObservableList
from pluginsmanager.observer.update_type import UpdateType

from unittest.mock import MagicMock


class PedalboardError(Exception):
    pass


class Pedalboard(object):
    """
    Pedalboard is a patch representation: your structure contains
    :class:`.Effect` and :class:`~pluginsmanager.model.connection.Connection`::

        >>> pedalboard = Pedalboard('Rocksmith')
        >>> bank.append(pedalboard)

        >>> builder = Lv2EffectBuilder()
        >>> pedalboard.effects
        ObservableList: []
        >>> reverb = builder.build('http://calf.sourceforge.net/plugins/Reverb')
        >>> pedalboard.append(reverb)
        >>> pedalboard.effects
        ObservableList: [<Lv2Effect object as 'Calf Reverb'  active at 0x7f60effb09e8>]

        >>> fuzz = builder.build('http://guitarix.sourceforge.net/plugins/gx_fuzzfacefm_#_fuzzfacefm_')
        >>> pedalboard.effects.append(fuzz)

        >>> pedalboard.connections
        ObservableList: []
        >>> pedalboard.connections.append(Connection(sys_effect.outputs[0], fuzz.inputs[0])) # View SystemEffect for more details
        >>> pedalboard.connections.append(Connection(fuzz.outputs[0], reverb.inputs[0]))
        >>> # It works too
        >>> pedalboard.connect(reverb.outputs[1], sys_effect.inputs[0])
        >>> pedalboard.connections
        ObservableList: [<Connection object as 'system.capture_1 -> GxFuzzFaceFullerMod.In' at 0x7f60f45f3f60>, <Connection object as 'GxFuzzFaceFullerMod.Out -> Calf Reverb.In L' at 0x7f60f45f57f0>, <Connection object as 'Calf Reverb.Out R -> system.playback_1' at 0x7f60f45dacc0>]

        >>> pedalboard.data
        {}
        >>> pedalboard.data = {'my-awesome-component': True}
        >>> pedalboard.data
        {'my-awesome-component': True}

    For load the pedalboard for play the songs with it::

        >>> mod_host.pedalboard = pedalboard

    All changes¹ in the pedalboard will be reproduced in mod-host.
    ¹ Except in data attribute, changes in this does not interfere with anything.

    :param string name: Pedalboard name
    """
    def __init__(self, name):
        self.name = name
        self._effects = ObservableList()
        self._connections = ObservableList()

        self.effects.observer = self._effects_observer
        self.connections.observer = self._connections_observer

        self._observer = MagicMock()

        self.bank = None

        self.data = {}

    @property
    def observer(self):
        return self._observer

    @observer.setter
    def observer(self, observer):
        self._observer = observer

        for effect in self.effects:
            effect.observer = observer

    def _effects_observer(self, update_type, effect, index, **kwargs):
        kwargs['index'] = index
        kwargs['origin'] = self

        if update_type == UpdateType.CREATED:
            self._init_effect(effect)

        elif update_type == UpdateType.UPDATED:
            self._init_effect(effect)

            old_effect = kwargs['old']
            if old_effect not in self.effects:
                self._clear_effect(old_effect)

        elif update_type == UpdateType.DELETED:
            self._clear_effect(effect)

        self.observer.on_effect_updated(effect, update_type, index=index, origin=self)

    def _init_effect(self, effect):
        effect.pedalboard = self
        effect.observer = self.observer

    def _clear_effect(self, effect):
        for connection in effect.connections:
            self.connections.real_list.remove(connection)

        effect.pedalboard = None
        effect.observer = MagicMock()

    def _connections_observer(self, update_type, connection, index, **kwargs):
        self.observer.on_connection_updated(connection, update_type, pedalboard=self, **kwargs)

    @property
    def json(self):
        """
        Get a json decodable representation of this pedalboard

        :return dict: json representation
        """
        return self.__dict__

    @property
    def __dict__(self):
        return {
            'name': self.name,
            'effects': [effect.json for effect in self.effects],
            'connections': [connection.json for connection in self.connections],
            'data': self.data
        }

    def append(self, effect):
        """
        Add a :class:`.Effect` in this pedalboard

        This works same as::

            >>> pedalboard.effects.append(effect)

        or::

            >>> pedalboard.effects.insert(len(pedalboard.effects), effect)

        :param Effect effect: Effect that will be added
        """
        if effect.is_unique_for_all_pedalboards:
            raise PedalboardError("The effect '{}' is unique for all pedalboards. "
                                  "Then, isn't allowed add it in any pedalboard.".format(str(effect)))

        self.effects.append(effect)

    @property
    def effects(self):
        """
        Return the effects presents in the pedalboard

        .. note::

            Because the effects is an :class:`.ObservableList`, it isn't settable.
            For replace, del the effects unnecessary and add the necessary
            effects
        """
        return self._effects

    @property
    def connections(self):
        """
        Return the pedalboard connections list

        .. note::

            Because the connections is an :class:`.ObservableList`, it isn't settable.
            For replace, del the connections unnecessary and add the necessary
            connections
        """
        return self._connections

    @property
    def index(self):
        """
        Returns the first occurrence of the pedalboard in your bank
        """
        if self.bank is None:
            raise IndexError('Pedalboard not contains a bank')

        return self.bank.pedalboards.index(self)

    def connect(self, output_port, input_port):
        """
        Connect two :class:`.Effect` instances in this pedalboard.
        For this, is necessary informs the output port origin and the input port destination::

            >>> pedalboard.append(driver)
            >>> pedalboard.append(reverb)
            >>> driver_output = driver.outputs[0]
            >>> reverb_input = reverb.inputs[0]
            >>> Connection(driver_output, reverb_input) in driver.connections
            False
            >>> pedalboard.connect(driver_output, reverb_input)
            >>> Connection(driver_output, reverb_input) in driver.connections
            True

        :param Port output_port: Effect output port
        :param Port input_port: Effect input port
        """
        ConnectionClass = output_port.connection_class
        self.connections.append(ConnectionClass(output_port, input_port))

    def disconnect(self, output_port, input_port):
        """
        Remove a connection between (two ports of) :class:`.Effect` instances.
        For this, is necessary informs the output port origin and the input port destination::

            >>> pedalboard.append(driver)
            >>> pedalboard.append(reverb)
            >>> driver_output = driver.outputs[0]
            >>> reverb_input = reverb.inputs[0]
            >>> pedalboard.connect(driver_output, reverb_input)
            >>> Connection(driver_output, reverb_input) in driver.connections
            True
            >>> pedalboard.disconnect(driver_output, reverb_input)
            >>> Connection(driver_output, reverb_input) in driver.connections
            False

        :param Port output_port: Effect output port
        :param Port input_port: Effect input port
        """
        ConnectionClass = output_port.connection_class
        self.connections.remove(ConnectionClass(output_port, input_port))



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.observer.update_type import UpdateType


class ObservableList(object):
    """
    Detects changes in list.

    In append, in remove and in setter, the `observer` is callable with changes
    details

    Based in https://www.pythonsheets.com/notes/python-basic.html#emulating-a-list
    """

    def __init__(self, lista=None):
        self.real_list = lista if lista is not None else []
        self.observer = lambda *args, **kwargs: ...

    def __str__(self):
        """
        See :meth:`list.__repr__()` method
        """
        return repr(self.real_list)

    def __repr__(self):
        """
        See :meth:`list.__repr__()` method
        """
        return "ObservableList: " + repr(self.real_list)

    def append(self, item):
        """
        See :meth:`list.append()` method

        Calls observer ``self.observer(UpdateType.CREATED, item, index)`` where
        **index** is *item position*
        """
        self.real_list.append(item)
        self.observer(UpdateType.CREATED, item, len(self.real_list) - 1)

    def remove(self, item):
        """
        See :meth:`list.remove()` method

        Calls observer ``self.observer(UpdateType.DELETED, item, index)`` where
        **index** is *item position*
        """
        index = self.index(item)
        self.real_list.remove(item)
        self.observer(UpdateType.DELETED, item, index)

    def index(self, x):
        """
        See :meth:`list.index()` method
        """
        return self.real_list.index(x)

    def insert(self, index, x):
        """
        See :meth:`list.insert()` method

        Calls observer ``self.observer(UpdateType.CREATED, item, index)``
        """
        self.real_list.insert(index, x)
        self.observer(UpdateType.CREATED, x, index)

    def pop(self, index=None):
        """
        See :meth:`list.pop()` method

        Remove the item at the given position in the list, and return it. If no index is specified,
        a.pop() removes and returns the last item in the list.

        :param int index: element index that will be removed

        :return: item removed
        """
        if index is None:
            index = len(self.real_list) - 1

        item = self[index]
        del self[index]

        return item

    def __len__(self):
        """
        See :meth:`list.__len__()` method
        """
        return len(self.real_list)

    def __getitem__(self, index):
        """
        See :meth:`list.__getitem__()` method
        """
        return self.real_list[index]

    def __setitem__(self, index, val):
        """
        See :meth:`list.__setitem__()` method

        Calls observer ``self.observer(UpdateType.UPDATED, item, index)``
        if ``val != self[index]``
        """
        if val == self[index]:
            return

        old = self.real_list[index]
        self.real_list[index] = val

        self.observer(UpdateType.UPDATED, val, index, old=old)

    def __delitem__(self, sliced):
        """
        See :meth:`list.__delitem__()` method

        Calls observer ``self.observer(UpdateType.DELETED, item, index)``
        where **item** is `self[index]`
        """
        item = self.real_list[sliced]
        del self.real_list[sliced]
        self.observer(UpdateType.DELETED, item, sliced)

    def __contains__(self, item):
        """
        See :meth:`list.__contains__()` method
        """
        return item in self.real_list

    def __iter__(self):
        """
        See :meth:`list.__iter__()` method
        """
        return iter(self.real_list)

    def move(self, item, new_position):
        """
        Moves a item list to new position

        Calls observer ``self.observer(UpdateType.DELETED, item, index)``
        and observer ``self.observer(UpdateType.CREATED, item, index)``
        if ``val != self[index]``

        :param item: Item that will be moved to new_position
        :param new_position: Item's new position
        """
        if item == self[new_position]:
            return

        self.remove(item)
        self.insert(new_position, item)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.system.system_input import SystemInput
from pluginsmanager.model.system.system_output import SystemOutput


class ProtocolParser:
    """
    Prepare the objects to `mod-host`_ string command

    .. _mod-host: https://github.com/moddevices/mod-host
    """

    @staticmethod
    def add(effect):
        """
        ``add <lv2_uri> <instance_number>``

        add a LV2 plugin encapsulated as a jack client

        e.g.::

            add http://lv2plug.in/plugins/eg-amp 0

        instance_number must be any value between 0 ~ 9999, inclusively

        :param Lv2Effect effect: Effect will be added
        """
        plugin = effect.plugin
        return "add {} {}".format(plugin['uri'], effect.instance)

    @staticmethod
    def remove(effect):
        """
        ``remove <instance_number>``

        remove a LV2 plugin instance (and also the jack client)

        e.g.::

            remove 0

        :param Lv2Effect effect: Effect will be removed
        """
        return 'remove {}'.format(effect.instance)

    @staticmethod
    def connect(connection):
        """
        ``connect <origin_port> <destination_port>``

        connect two plugin audio ports

        e.g.::

            connect system:capture_1 plugin_0:in

        :param pluginsmanager.model.connection.Connection connection: Connection with a valid
               :class:`.Output` and :class:`.Input`
        """
        return ProtocolParser._connect_message(
            ProtocolParser._get_out_name_of(connection.output),
            ProtocolParser._get_in_name_of(connection.input)
        )

    @staticmethod
    def _connect_message(origin_port, destination_port):
        return 'connect {} {}'.format(origin_port, destination_port)

    @staticmethod
    def _get_out_name_of(effect_output):
        effect = effect_output.effect

        if isinstance(effect_output, SystemOutput):
            return '{}:{}'.format(effect, effect_output)

        return 'effect_{}:{}'.format(effect.instance, effect_output.symbol)

    @staticmethod
    def _get_in_name_of(effect_input):
        effect = effect_input.effect

        if isinstance(effect_input, SystemInput):
            return '{}:{}'.format(effect, effect_input)

        return 'effect_{}:{}'.format(effect.instance, effect_input.symbol)

    @staticmethod
    def disconnect(connection):
        """
        ``disconnect <origin_port> <destination_port>``

        disconnect two plugin audio ports

        e.g.::

            disconnect system:capture_1 plugin_0:in

        :param pluginsmanager.model.connection.Connection connection: Connection with a valid
               :class:`.Output` and :class:`.Input`
        """
        return 'disconnect {} {}'.format(
            ProtocolParser._get_out_name_of(connection.output),
            ProtocolParser._get_in_name_of(connection.input)
        )

    @staticmethod
    def preset_load():
        """
        ``preset_load <instance_number> <preset_uri>``

        load a preset state to given plugin instance

        e.g.::

            preset_load 0 "http://drobilla.net/plugins/mda/presets#JX10-moogcury-lite"

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def preset_save():
        """
        ``preset_save <instance_number> <preset_name> <dir> <file_name>``

        save a preset state from given plugin instance

        e.g.::

            preset_save 0 "My Preset" /home/user/.lv2/my-presets.lv2 mypreset.ttl

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def preset_show():
        """
        ``preset_show <instance_number> <preset_uri>``

        show the preset information of requested instance / URI

        e.g.::

            preset_show 0 http://drobilla.net/plugins/mda/presets#EPiano-bright

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def param_set(param):
        """
        ``param_set <instance_number> <param_symbol> <param_value>``

        set a value to given control

        e.g.::

            param_set 0 gain 2.50

        :param Lv2Param param: Parameter that will be updated your value
        """
        instance = param.effect.instance

        return 'param_set {} {} {}'.format(instance, param.symbol, param.value)

    @staticmethod
    def param_get(param):
        """
        ``param_get <instance_number> <param_symbol>``

        get the value of the request control

        e.g.::

            param_get 0 gain

        :param Lv2Param param: Parameter that will be get your current value
        """
        instance = param.effect.instance

        return 'param_get {} {}'.format(instance, param.symbol)

    @staticmethod
    def param_monitor():
        """
        ``param_monitor <instance_number> <param_symbol> <cond_op> <value>``

        do monitoring a plugin instance control port according given condition

        e.g.::

            param_monitor 0 gain > 2.50

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def monitor():
        """
        ``monitor <addr> <port> <status>``

        open a socket port to monitoring parameters

        e.g.::

            monitor localhost 12345 1

        * if ``status = 1`` start monitoring
        * if ``status = 0`` stop monitoring

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def midi_learn(self, plugin, param):
        """
        ``midi_learn <instance_number> <param_symbol>``

        This command maps starts MIDI learn for a parameter

        e.g.::

            midi_learn 0 gain

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def midi_map(plugin, param, midi_chanel, midi_cc):
        """
        ``midi_map <instance_number> <param_symbol> <midi_channel> <midi_cc>``

        This command maps a MIDI controller to a parameter

        e.g.::

            midi_map 0 gain 0 7

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def midi_unmap(plugin, param):
        """
        ``midi_unmap <instance_number> <param_symbol>``

        This command unmaps the MIDI controller from a parameter

        e.g.::

            unmap 0 gain

        .. note::

            Not implemented yet
        """
        pass

    @staticmethod
    def bypass(effect):
        """
        ``bypass <instance_number> <bypass_value>``

        toggle plugin processing

        e.g.::

            bypass 0 1

        * if ``bypass_value = 1`` bypass plugin
        * if ``bypass_value = 0`` process plugin

        :param Lv2Effect effect: Effect that will be active the bypass
               or disable the bypass
        """
        return 'bypass {} {}'.format(
            effect.instance,
            1 if effect.active else 0
        )

    @staticmethod
    def load(filename):
        """
        ``load <file_name>``

        load a history command file
        dummy way to save/load workspace state

        e.g.::

            load my_setup

        .. note::

            Not implemented yet
        """
        return 'load {}'.format(filename)

    @staticmethod
    def save(filename):
        """
        ``save <file_name>``

        saves the history of typed commands
        dummy way to save/load workspace state

        e.g.::

            save my_setup

        .. note::

            Not implemented yet
        """
        return 'save {}'.format(filename)

    @staticmethod
    def help():
        """
        ``help``

        show a help message
        """
        return 'help'

    @staticmethod
    def quit():
        """
        ``quit``

        bye!
        """
        return 'quit'




# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.effect import Effect
from pluginsmanager.model.lv2.lv2_param import Lv2Param
from pluginsmanager.model.lv2.lv2_input import Lv2Input
from pluginsmanager.model.lv2.lv2_output import Lv2Output
from pluginsmanager.model.lv2.lv2_midi_input import Lv2MidiInput
from pluginsmanager.model.lv2.lv2_midi_output import Lv2MidiOutput

from pluginsmanager.util.dict_tuple import DictTuple


class Lv2Effect(Effect):
    """
    Representation of a Lv2 audio plugin instance.

    For general effect use, see :class:`.Effect` class documentation.

    It's possible obtains the :class:`.Lv2Plugin` information::

        >>> reverb
        <Lv2Effect object as 'Calf Reverb'  active at 0x7f60effb09e8>
        >>> reverb.plugin
        <Lv2Plugin object as Calf Reverb at 0x7f60effb9940>

    :param Lv2Plugin plugin:
    """

    def __init__(self, plugin):
        super(Lv2Effect, self).__init__()

        self.plugin = plugin

        params = [Lv2Param(self, param) for param in plugin["ports"]["control"]["input"]]
        self._params = DictTuple(params, lambda param: param.symbol)

        inputs = [Lv2Input(self, effect_input) for effect_input in plugin['ports']['audio']['input']]
        self._inputs = DictTuple(inputs, lambda _input: _input.symbol)

        outputs = [Lv2Output(self, effect_output) for effect_output in plugin['ports']['audio']['output']]
        self._outputs = DictTuple(outputs, lambda _output: _output.symbol)

        midi_inputs = [Lv2MidiInput(self, midi_input) for midi_input in plugin['ports']['midi']['input']]
        self._midi_inputs = DictTuple(midi_inputs, lambda _output: _output.symbol)

        midi_outputs = [Lv2MidiOutput(self, midi_output) for midi_output in plugin['ports']['midi']['output']]
        self._midi_outputs = DictTuple(midi_outputs, lambda _output: _output.symbol)

        self.instance = None

    def __str__(self):
        return str(self.plugin)

    def __repr__(self):
        return "<{} object as '{}' {} active at 0x{:x}>".format(
            self.__class__.__name__,
            str(self),
            '' if self.active else 'not',
            id(self)
        )

    @property
    def __dict__(self):
        return {
            'technology': 'lv2',
            'plugin': self.plugin['uri'],
            'active': self.active,
            'params': [param.json for param in self.params],
            'version': self.version
        }

    @property
    def version(self):
        """
        :return string: Version of plugin of effect
        """
        return self.plugin.version



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import os


class Persistence(object):

    @staticmethod
    def read(path, create_file=False):
        """
        Reads the json data in path

        :param Path path: Path that json data will be readed
        :param create_file: Creates the file if it isn't exists

        :return: json data
        """
        if create_file:
            with open(str(path), 'a+') as data_file:
                data_file.seek(0)
                return json.load(data_file)

        else:
            with open(str(path)) as data_file:
                return json.load(data_file)

    @staticmethod
    def save(path, json_data):
        """
        Saves json_data in path

        :param Path path: Path that json_data will be persisted
        :param json_data: Data that will be persisted
        """
        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(Persistence._save(path, json_data))
        except AssertionError:
            Persistence._save(path, json_data)

    @staticmethod
    @asyncio.coroutine
    def _save(path, json_data):
        with open(str(path), "w+") as file:
            file.write(json.dumps(json_data))

    @staticmethod
    def delete(path):
        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(Persistence._delete(path))
        except AssertionError:
            Persistence._delete(str(path))

    @staticmethod
    @asyncio.coroutine
    def _delete(path):
        os.remove(str(path))



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.input import Input
from pluginsmanager.model.lv2.lv2_port_mixing import Lv2PortMixing


class Lv2Input(Lv2PortMixing, Input):
    """
    Representation of a Lv2 `input audio port`_ instance.

    For general input use, see :class:`.Input` class documentation.

    .. _input audio port: http://lv2plug.in/ns/lv2core/#InputPort

    :param Lv2Effect effect: Effect that contains the input
    :param dict data: *input audio port* json representation
    """

    def __init__(self, effect, data):
        super(Lv2Input, self).__init__(effect)
        self._data = data

    @property
    def data(self):
        return self._data



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import subprocess

from pluginsmanager.model.lv2.lv2_plugin import Lv2Plugin
from pluginsmanager.model.lv2.lv2_effect import Lv2Effect


class Lv2EffectBuilder(object):
    """
    Generates lv2 audio plugins instance (as :class:`.Lv2Effect` object).

    .. note::

        In the current implementation, the data plugins are persisted
        in *plugins.json*.

    :param Path plugins_json: Plugins json path file
    """

    plugins_json_file = os.path.dirname(os.path.abspath(__file__)) + '/plugins.json'
    """
    Informs the path of the `plugins.json` file. This file contains the lv2 plugins metadata info
    """

    def __init__(self, plugins_json=None):
        self._plugins = {}

        if plugins_json is None:
            plugins_json = Lv2EffectBuilder.plugins_json_file

        with open(str(plugins_json)) as data_file:
            data = json.load(data_file)

        self.reload(data)

    def reload(self, metadata):
        """
        Loads the metadata. They will be used so that it is possible to generate lv2 audio plugins.

        :param list metadata: lv2 audio plugins metadata
        """
        supported_plugins = self._supported_plugins
        for plugin in metadata:
            if plugin['uri'] in supported_plugins:
                self._plugins[plugin['uri']] = Lv2Plugin(plugin)

    @property
    def _supported_plugins(self):
        return str(subprocess.check_output(['lv2ls'])).split('\\n')

    @property
    def all(self):
        return self._plugins

    @property
    def plugins(self):
        return self._plugins.keys()

    def build(self, lv2_uri):
        """
        Returns a new :class:`.Lv2Effect` by the valid lv2_uri

        :param string lv2_uri:
        :return Lv2Effect: Effect created
        """
        return Lv2Effect(self._plugins[lv2_uri])

    def lv2_plugins_data(self):
        """
        Generates a file with all plugins data info. It uses the `lilvlib`_ library.

        PluginsManager can manage lv2 audio plugins through previously obtained metadata
        from the lv2 audio plugins descriptor files.

        To speed up usage, data has been pre-generated and loaded into this piped packet.
        This avoids a dependency installation in order to obtain the metadata.

        However, this measure makes it not possible to manage audio plugins that were not
        included in the list.

        To work around this problem, this method - using the `lilvlib`_ library - can get
        the information from the audio plugins. You can use this data to generate a file
        containing the settings::

            >>> builder = Lv2EffectBuilder()
            >>> plugins_data = builder.lv2_plugins_data()

            >>> import json
            >>> with open('plugins.json', 'w') as outfile:
            >>>     json.dump(plugins_data, outfile)

        The next time you instantiate this class, you can pass the configuration file::

            >>> builder = Lv2EffectBuilder(os.path.abspath('plugins.json'))

        Or, if you want to load the data without having to create a new instance of this class::

            >>> builder.reload(builder.lv2_plugins_data())

        .. warning::

            To use this method, it is necessary that the system has the `lilv`_ in a version equal
            to or greater than `0.22.0`_. Many linux systems currently have previous versions on
            their package lists, so you need to compile them manually.

            In order to ease the work, Pedal Pi has compiled lilv for some versions of linux.
            You can get the list of .deb packages in https://github.com/PedalPi/lilvlib/releases.

            .. code-block:: bash

                # Example
                wget https://github.com/PedalPi/lilvlib/releases/download/v1.0.0/python3-lilv_0.22.1.git20160613_amd64.deb
                sudo dpkg -i python3-lilv_0.22.1+git20160613_amd64.deb


            If the architecture of your computer is not contemplated, moddevices provided
            a script to generate the package.
            Go to https://github.com/moddevices/lilvlib to get the script in its most up-to-date version.

        .. _lilvlib: https://github.com/moddevices/lilvlib
        .. _0.22.0: http://git.drobilla.net/cgit.cgi/lilv.git/tag/?id=v0.22.0
        .. _lilv: http://drobilla.net/software/lilv

        :return list: lv2 audio plugins metadata
        """
        import lilvlib

        return lilvlib.get_plugin_info_helper('')

if __name__ == '__main__':
    builder = Lv2EffectBuilder()
    print('Total plugins before reload:', len(builder.plugins))

    builder.reload(builder.lv2_plugins_data())
    print('Total plugins after reload:', len(builder.plugins))



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from pluginsmanager.banks_manager import BanksManager
from pluginsmanager.observer.autosaver.banks_files import BanksFiles
from pluginsmanager.observer.autosaver.index_file import IndexFile
from pluginsmanager.observer.update_type import UpdateType
from pluginsmanager.observer.updates_observer import UpdatesObserver


class Autosaver(UpdatesObserver):
    """
    The UpdatesObserver :class:`.Autosaver` allows save any changes
    automatically in json data files.
    Save all plugins changes in json files in a specified path.

    It also allows loading of saved files::

        >>> system_effect = SystemEffect('system', ('capture_1', 'capture_2'), ('playback_1', 'playback_2'))
        >>>
        >>> autosaver = Autosaver('my/path/data/')
        >>> banks_manager = autosaver.load(system_effect)

    When loads data with :class:`.Autosaver`, the autosaver
    has registered in observers of the banks_manager generated::

        >>> autosaver in banks_manager.observers
        True

    For manual registering in :class:`.BanksManager` uses
    :func:`~pluginsmanager.banks_manager.BanksManager.register()`::

        >>> banks_manager = BanksManager()
        >>> autosaver = Autosaver('my/path/data/')
        >>> autosaver in banks_manager.observers
        False
        >>> banks_manager.register(autosaver)
        >>> autosaver in banks_manager.observers
        True

    After registered, any changes in :class:`.Bank`, :class:`.Pedalboard`, :class:`.Effect`,
    :class:`~pluginsmanager.model.connection.Connection` or :class:`.Param` which belong
    to the structure of :class:`.BanksManager` instance are persisted automatically
    by :class:`.Autosaver`::

        >>> banks_manager = BanksManager()
        >>> banks_manager.register(autosaver)
        >>> my_bank = Bank('My bank')
        >>> banks_manager.append(my_bank)
        >>> # The bank will be added in banksmanger
        >>> # and now is observable (and persisted) by autosaver

    It's possible disables autosaver for saves manually::

        >>> autosaver.auto_save = False
        >>> autosaver.save(banks_manager)  # save() method saves all banks data

    :param string data_path: Path that banks will be saved (each bank in one file)
    :param bool auto_save: Auto save any change?
    """
    def __init__(self, data_path, auto_save=True):
        super().__init__()
        self.data_path = Path(data_path)

        self.index_file = IndexFile(self.data_path / Path('index_file'))
        self.banks_files = BanksFiles(self.data_path)

        self.auto_save = auto_save

    def load(self, system_effect):
        """
        Return a :class:`.BanksManager` instance contains the banks present in
        :attr:`~pluginsmanager.observer.autosaver.autosaver.Autosaver.data_path`

        :param SystemEffect system_effect: SystemEffect used in pedalboards
        :return BanksManager: :class:`.BanksManager` with banks persisted in
                :attr:`~pluginsmanager.observer.autosaver.autosaver.Autosaver.data_path`
        """
        banks = self.banks_files.load(system_effect)
        banks_ordered = self.index_file.load(banks)

        manager = BanksManager()
        manager.register(self)

        for bank in banks_ordered:
            manager.append(bank)
            bank.manager = manager

        return manager

    def save(self, banks_manager):
        """
        Save all data from a banks_manager

        :param BanksManager banks_manager: BanksManager that your banks data will be persisted
        """
        self.banks_files.delete_all_banks()
        self.banks_files.save(banks_manager)
        self.index_file.save(banks_manager)

    def on_bank_updated(self, bank, update_type, index, origin, **kwargs):
        if not self.auto_save:
            return

        if update_type == UpdateType.DELETED:
            self.banks_files.delete_bank(bank)

        elif update_type == UpdateType.CREATED:
            self.banks_files.save_bank(bank)

        elif update_type == UpdateType.UPDATED:
            self.banks_files.save_bank(bank)
            old_bank = kwargs['old']

            if old_bank.manager is None:
                self.banks_files.delete_bank(old_bank)

        self.index_file.save(origin)

    def on_pedalboard_updated(self, pedalboard, update_type, index, origin, **kwargs):
        if not self.auto_save:
            return

        if update_type == UpdateType.DELETED:
            self.banks_files.save_bank(origin)
        else:
            self.banks_files.save_bank(pedalboard.bank)

    def on_effect_updated(self, effect, update_type, index, origin, **kwargs):
        if not self.auto_save:
            return

        pedalboard = origin
        self.banks_files.save_bank(pedalboard.bank)

    def on_effect_status_toggled(self, effect, **kwargs):
        if not self.auto_save:
            return

        self.banks_files.save_bank(effect.pedalboard.bank)

    def on_param_value_changed(self, param, **kwargs):
        if not self.auto_save:
            return

        self.banks_files.save_bank(param.effect.pedalboard.bank)

    def on_connection_updated(self, connection, update_type, pedalboard, **kwargs):
        if not self.auto_save:
            return

        self.banks_files.save_bank(pedalboard.bank)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.util.builder.builder import AudioPortBuilder


class SystemAudioPortBuilder(AudioPortBuilder):
    """
    Extracts the :class:`.SystemInput`s and :class:`.SystemOutputs`s of an :class:`.SystemEffect` defined in a json.
    """

    def __init__(self, system_effect):
        self.system_effect = system_effect

    def build_input(self, json):
        symbol = json['symbol']
        return self.system_effect.inputs[symbol]

    def build_output(self, json):
        symbol = json['symbol']
        return self.system_effect.outputs[symbol]

    def build_midi_input(self, json):
        symbol = json['symbol']
        return self.system_effect.midi_inputs[symbol]

    def build_midi_output(self, json):
        symbol = json['symbol']
        return self.system_effect.midi_outputs[symbol]



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess

from pluginsmanager.observer.updates_observer import UpdatesObserver
from pluginsmanager.observer.update_type import UpdateType

from pluginsmanager.observer.mod_host.host import Host
from pluginsmanager.util.pairs_list import PairsList


class ModHostError(Exception):

    def __init__(self, *args, **kwargs):
        super(ModHostError, self).__init__(*args, **kwargs)


class ModHost(UpdatesObserver):
    """
    **Python port for mod-host**
        `Mod-host`_ is a `LV2`_ host for Jack controllable via socket or command line.

    This class offers the mod-host control in a python API::

        # Create a mod-host, connect and register it in banks_manager
        mod_host = ModHost('localhost')
        mod_host.connect()
        banks_manager.register(mod_host)

        # Set the mod_host pedalboard for a pedalboard that the bank
        # has added in banks_manager
        mod_host.pedalboard = my_awesome_pedalboard

    The changes in current pedalboard (:attr:`~pluginsmanager.mod_host.mod_host.ModHost.pedalboard`
    attribute of `mod_host`) will also result in mod-host::

        driver = my_awesome_pedalboard.effects[0]
        driver.active = False

    .. note::

        For use, is necessary that the mod-host is running, for use, access

         * `Install dependencies`_
         * `Building mod-host`_
         * `Running mod-host`_

        For more JACK information, access `Demystifying JACK – A Beginners Guide to Getting Started with JACK`_

    **Example:**

        In this example, is starting a `Zoom G3`_ series audio interface. Others interfaces
        maybe needs others configurations.

    .. code-block:: bash

        # Starting jackdump process via console
        jackd -R -P70 -t2000 -dalsa -dhw:Series -p256 -n3 -r44100 -s &
        # Starting mod-host
        mod-host &

    :param string address: Computer mod-host process address (IP). If the
     process is running on the same computer that is running the python code
     uses `localhost`.
    :param int port: Socket port on which mod-host should be running. Default is `5555`

    .. _Mod-host: https://github.com/moddevices/mod-host
    .. _LV2: http://lv2plug.in
    .. _Install dependencies: https://github.com/deedos/mod-host/commit/0941d84fc48deb74e27cdcbf23a88db2007d5c6f
    .. _Zoom G3: https://www.zoom.co.jp/products/guitar/g3-guitar-effects-amp-simulator-pedal
    .. _Building mod-host: https://github.com/moddevices/mod-host#building
    .. _Running mod-host: https://github.com/moddevices/mod-host#running
    .. _Demystifying JACK – A Beginners Guide to Getting Started with JACK: http://libremusicproduction.com/articles/demystifying-jack-%E2%80%93-beginners-guide-getting-started-jack
    """

    def __init__(self, address='localhost', port=5555):
        super(ModHost, self).__init__()
        self.address = address
        self.port = port

        self.process = 'mod-host'

        self.host = None
        self._pedalboard = None

        self.pairs_list = PairsList(lambda effect: effect.plugin['uri'])

        self._started_with_this_api = False

    def start(self):
        """
        Invokes the mod-host process.

        mod-host requires JACK to be running.
        mod-host does not startup JACK automatically, so you need to start it before running mod-host.

        .. note::

            This function is experimental. There is no guarantee that the process will actually be initiated.
        """
        if self.address != 'localhost':
            raise ModHostError('The host configured in the constructor isn''t "localhost". '
                               'It is not possible to start a process on another device.')

        try:
            subprocess.call([self.process, '-p', str(self.port)])

        except FileNotFoundError as e:
            exception = ModHostError(
                'mod-host not found. Did you install it? '
                '(https://github.com/moddevices/mod-host#building)'
            )

            raise exception from e

        self._started_with_this_api = True

    def connect(self):
        """
        Connect the object with mod-host with the _address_ parameter informed in
        the constructor method (:meth:`~pluginsmanager.mod_host.mod_host.ModHost.__init__()`)
        """
        self.host = Host(self.address, self.port)

    @property
    def pedalboard(self):
        """
        Currently managed pedalboard (current pedalboard)

        :getter: Current pedalboard - Pedalboard loaded by mod-host
        :setter: Set the pedalboard that will be loaded by mod-host
        :type: Pedalboard
        """
        return self._pedalboard

    @pedalboard.setter
    def pedalboard(self, pedalboard):
        self.on_current_pedalboard_changed(pedalboard)

    def __del__(self):
        """
        Calls :meth:`~pluginsmanager.mod_host.mod_host.ModHost.close()` method for
        remove the audio plugins loaded and closes connection
        with mod-host.

            >>> mod_host = ModHost()
            >>> del mod_host

        .. note::

            If the mod-host process has been created with :meth:`~pluginsmanager.mod_host.mod_host.ModHost.start()`
            method, it will be finished.
        """
        self.close()

    def close(self):
        """
        Remove the audio plugins loaded and closes connection with mod-host.

        .. note::

            If the mod-host process has been created with :meth:`~pluginsmanager.mod_host.mod_host.ModHost.start()`
            method, it will be finished.
        """
        if self.host is None:
            raise ModHostError('There is no established connection with mod-host. '
                               'Did you call the `connect()` method?')
        self.pedalboard = None

        if self._started_with_this_api:
            self.host.quit()
        else:
            self.host.close()

    ####################################
    # Observer
    ####################################
    def on_current_pedalboard_changed(self, pedalboard, **kwargs):
        if self.pedalboard is not None and pedalboard is not None:
            self._replace_pedalboard(self.pedalboard, pedalboard)
        else:
            self._change_pedalboard(pedalboard)

    def on_bank_updated(self, bank, update_type, **kwargs):
        if (self.pedalboard is not None
        and bank != self.pedalboard.bank):
            return
        pass

    def on_pedalboard_updated(self, pedalboard, update_type, **kwargs):
        if pedalboard != self.pedalboard:
            return

        self.on_current_pedalboard_changed(pedalboard)

    def on_effect_updated(self, effect, update_type, index, origin, **kwargs):
        if origin != self.pedalboard:
            return

        if update_type == UpdateType.CREATED:
            self.host.add(effect)
            self._load_params_of(effect)
            self.on_effect_status_toggled(effect)

        if update_type == UpdateType.DELETED:
            self.host.remove(effect)

    def _load_params_of(self, effect):
        """
        Called only when a effect has created
        Param changes calls :meth:`~pluginsmanager.mod_host.mod_host.ModHost.on_param_value_changed()`
        """
        for param in effect.params:
            if param.value != param.default:
                self._set_param_value(param)

    def on_effect_status_toggled(self, effect, **kwargs):
        if effect.pedalboard != self.pedalboard:
            return

        self.host.set_status(effect)

    def on_param_value_changed(self, param, **kwargs):
        if param.effect.pedalboard != self.pedalboard:
            return

        self._set_param_value(param)

    def on_connection_updated(self, connection, update_type, pedalboard, **kwargs):
        if pedalboard != self.pedalboard:
            return

        if update_type == UpdateType.CREATED:
            self.host.connect(connection)
        elif update_type == UpdateType.DELETED:
            self.host.disconnect(connection)

    def _set_param_value(self, param):
        self.host.set_param_value(param)

    ####################################
    # Private methods
    ####################################
    def _replace_pedalboard(self, current, pedalboard):
        # Replace effects with equal plugins
        result = self.pairs_list.calculate(current.effects, pedalboard.effects)

        for current_effect, new_effect in result.pairs:
            new_effect.instance = current_effect.instance

            for parameter_old_effect, parameter_new_effect in zip(current_effect.params, new_effect.params):
                if parameter_new_effect.value != parameter_old_effect.value:
                    self._set_param_value(parameter_new_effect)

        # Remove not equal plugins
        current_will_remove = result.elements_not_added_a

        self._remove_connections_of(current)
        self._remove_effects(current_will_remove)

        self._pedalboard = pedalboard

        # Remove not equal plugins
        # Changes are only updated if self._pedalboard = pedalboard
        pedalboard_will_add = result.elements_not_added_b

        self._add_effects(pedalboard_will_add)
        self._add_connections_of(pedalboard)

    def _change_pedalboard(self, pedalboard):
        if self.pedalboard is not None:
            self._remove_pedalboard(self.pedalboard)

        self._pedalboard = pedalboard

        # Changes are only updated if self._pedalboard = pedalboard
        if pedalboard is not None:
            self._add_effects(pedalboard.effects)
            self._add_connections_of(pedalboard)

    def _remove_pedalboard(self, pedalboard):
        self._remove_effects(pedalboard.effects)
        self._remove_connections_of(pedalboard)

    def _remove_connections_of(self, pedalboard):
        for connection in pedalboard.connections:
            self.on_connection_updated(connection, UpdateType.DELETED, pedalboard=pedalboard)

    def _remove_effects(self, effects):
        for effect in effects:
            self.on_effect_updated(effect, UpdateType.DELETED, index=None, origin=effect.pedalboard)

    def _add_connections_of(self, pedalboard):
        for connection in pedalboard.connections:
            self.on_connection_updated(connection, UpdateType.CREATED, pedalboard=pedalboard)

    def _add_effects(self, effects):
        for effect in effects:
            self.on_effect_updated(effect, UpdateType.CREATED, index=None, origin=effect.pedalboard)




"""
File copy from https://github.com/moddevices/lilvlib
"""
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------------------------------------
# Imports

import json
import lilv
import os

from math import fmod

# ------------------------------------------------------------------------------------------------------------
# Utilities

def LILV_FOREACH(collection, func):
    itr = collection.begin()
    while itr:
        yield func(collection.get(itr))
        itr = collection.next(itr)

class NS(object):
    def __init__(self, world, base):
        self.world = world
        self.base = base
        self._cache = {}

    def __getattr__(self, attr):
        if attr.endswith("_"):
            attr = attr[:-1]
        if attr not in self._cache:
            self._cache[attr] = lilv.Node(self.world.new_uri(self.base+attr))
        return self._cache[attr]

def is_integer(string):
    return string.strip().lstrip("-+").isdigit()

def get_short_port_name(portName):
    if len(portName) <= 16:
        return portName

    portName = portName.split("/",1)[0].split(" (",1)[0].split(" [",1)[0].strip()

    # cut stuff if too big
    if len(portName) > 16:
        portName = portName[0] + portName[1:].replace("a","").replace("e","").replace("i","").replace("o","").replace("u","")

        if len(portName) > 16:
            portName = portName[:16]

    return portName.strip()

# ------------------------------------------------------------------------------------------------------------

def get_category(nodes):
    lv2_category_indexes = {
        'DelayPlugin': ['Delay'],
        'DistortionPlugin': ['Distortion'],
        'WaveshaperPlugin': ['Distortion', 'Waveshaper'],
        'DynamicsPlugin': ['Dynamics'],
        'AmplifierPlugin': ['Dynamics', 'Amplifier'],
        'CompressorPlugin': ['Dynamics', 'Compressor'],
        'ExpanderPlugin': ['Dynamics', 'Expander'],
        'GatePlugin': ['Dynamics', 'Gate'],
        'LimiterPlugin': ['Dynamics', 'Limiter'],
        'FilterPlugin': ['Filter'],
        'AllpassPlugin': ['Filter', 'Allpass'],
        'BandpassPlugin': ['Filter', 'Bandpass'],
        'CombPlugin': ['Filter', 'Comb'],
        'EQPlugin': ['Filter', 'Equaliser'],
        'MultiEQPlugin': ['Filter', 'Equaliser', 'Multiband'],
        'ParaEQPlugin': ['Filter', 'Equaliser', 'Parametric'],
        'HighpassPlugin': ['Filter', 'Highpass'],
        'LowpassPlugin': ['Filter', 'Lowpass'],
        'GeneratorPlugin': ['Generator'],
        'ConstantPlugin': ['Generator', 'Constant'],
        'InstrumentPlugin': ['Generator', 'Instrument'],
        'OscillatorPlugin': ['Generator', 'Oscillator'],
        'ModulatorPlugin': ['Modulator'],
        'ChorusPlugin': ['Modulator', 'Chorus'],
        'FlangerPlugin': ['Modulator', 'Flanger'],
        'PhaserPlugin': ['Modulator', 'Phaser'],
        'ReverbPlugin': ['Reverb'],
        'SimulatorPlugin': ['Simulator'],
        'SpatialPlugin': ['Spatial'],
        'SpectralPlugin': ['Spectral'],
        'PitchPlugin': ['Spectral', 'Pitch Shifter'],
        'UtilityPlugin': ['Utility'],
        'AnalyserPlugin': ['Utility', 'Analyser'],
        'ConverterPlugin': ['Utility', 'Converter'],
        'FunctionPlugin': ['Utility', 'Function'],
        'MixerPlugin': ['Utility', 'Mixer'],
        #'MIDIPlugin': ['MIDI', 'Utility'],
    }
    mod_category_indexes = {
        'DelayPlugin': ['Delay'],
        'DistortionPlugin': ['Distortion'],
        'DynamicsPlugin': ['Dynamics'],
        'FilterPlugin': ['Filter'],
        'GeneratorPlugin': ['Generator'],
        'ModulatorPlugin': ['Modulator'],
        'ReverbPlugin': ['Reverb'],
        'SimulatorPlugin': ['Simulator'],
        'SpatialPlugin': ['Spatial'],
        'SpectralPlugin': ['Spectral'],
        'UtilityPlugin': ['Utility'],
        'MIDIPlugin': ['Utility', 'MIDI'],
    }

    def fill_in_lv2_category(node):
        category = node.as_string().replace("http://lv2plug.in/ns/lv2core#","")
        if category in lv2_category_indexes.keys():
            return lv2_category_indexes[category]
        return []

    def fill_in_mod_category(node):
        category = node.as_string().replace("http://moddevices.com/ns/mod#","")
        if category in mod_category_indexes.keys():
            return mod_category_indexes[category]
        return []

    categories = []
    for cat in [cat for catlist in LILV_FOREACH(nodes, fill_in_mod_category) for cat in catlist]:
        if cat not in categories:
            categories.append(cat)

    if len(categories) > 0:
        return categories

    for cat in [cat for catlist in LILV_FOREACH(nodes, fill_in_lv2_category) for cat in catlist]:
        if cat not in categories:
            categories.append(cat)

    return categories

def get_port_data(port, subj):
    nodes = port.get_value(subj.me)
    data  = []

    it = lilv.lilv_nodes_begin(nodes)
    while not lilv.lilv_nodes_is_end(nodes, it):
        dat = lilv.lilv_nodes_get(nodes, it)
        it  = lilv.lilv_nodes_next(nodes, it)
        if dat is None:
            continue
        data.append(lilv.lilv_node_as_string(dat))

    return data

def get_port_unit(miniuri):
  # using label, render, symbol
  units = {
      's': ["seconds", "%f s", "s"],
      'ms': ["milliseconds", "%f ms", "ms"],
      'min': ["minutes", "%f mins", "min"],
      'bar': ["bars", "%f bars", "bars"],
      'beat': ["beats", "%f beats", "beats"],
      'frame': ["audio frames", "%f frames", "frames"],
      'm': ["metres", "%f m", "m"],
      'cm': ["centimetres", "%f cm", "cm"],
      'mm': ["millimetres", "%f mm", "mm"],
      'km': ["kilometres", "%f km", "km"],
      'inch': ["inches", """%f\"""", "in"],
      'mile': ["miles", "%f mi", "mi"],
      'db': ["decibels", "%f dB", "dB"],
      'pc': ["percent", "%f%%", "%"],
      'coef': ["coefficient", "* %f", "*"],
      'hz': ["hertz", "%f Hz", "Hz"],
      'khz': ["kilohertz", "%f kHz", "kHz"],
      'mhz': ["megahertz", "%f MHz", "MHz"],
      'bpm': ["beats per minute", "%f BPM", "BPM"],
      'oct': ["octaves", "%f octaves", "oct"],
      'cent': ["cents", "%f ct", "ct"],
      'semitone12TET': ["semitones", "%f semi", "semi"],
      'degree': ["degrees", "%f deg", "deg"],
      'midiNote': ["MIDI note", "MIDI note %d", "note"],
  }
  if miniuri in units.keys():
      return units[miniuri]
  return ("","","")

# ------------------------------------------------------------------------------------------------------------
# get_bundle_dirname

def get_bundle_dirname(bundleuri):
    bundle = lilv.lilv_uri_to_path(bundleuri)

    if not os.path.exists(bundle):
        raise IOError(bundleuri)
    if os.path.isfile(bundle):
        bundle = os.path.dirname(bundle)

    return bundle

# ------------------------------------------------------------------------------------------------------------
# get_pedalboard_info

# Get info from an lv2 bundle
# @a bundle is a string, consisting of a directory in the filesystem (absolute pathname).
def get_pedalboard_info(bundle):
    # lilv wants the last character as the separator
    bundle = os.path.abspath(bundle)
    if not bundle.endswith(os.sep):
        bundle += os.sep

    # Create our own unique lilv world
    # We'll load a single bundle and get all plugins from it
    world = lilv.World()

    # this is needed when loading specific bundles instead of load_all
    # (these functions are not exposed via World yet)
    lilv.lilv_world_load_specifications(world.me)
    lilv.lilv_world_load_plugin_classes(world.me)

    # convert bundle string into a lilv node
    bundlenode = lilv.lilv_new_file_uri(world.me, None, bundle)

    # load the bundle
    world.load_bundle(bundlenode)

    # free bundlenode, no longer needed
    lilv.lilv_node_free(bundlenode)

    # get all plugins in the bundle
    plugins = world.get_all_plugins()

    # make sure the bundle includes 1 and only 1 plugin (the pedalboard)
    if plugins.size() != 1:
        raise Exception('get_pedalboard_info(%s) - bundle has 0 or > 1 plugin'.format(bundle))

    # no indexing in python-lilv yet, just get the first item
    plugin = None
    for p in plugins:
        plugin = p
        break

    if plugin is None:
        raise Exception('get_pedalboard_info(%s) - failed to get plugin, you are using an old lilv!'.format(bundle))

    # define the needed stuff
    ns_rdf      = NS(world, lilv.LILV_NS_RDF)
    ns_lv2core  = NS(world, lilv.LILV_NS_LV2)
    ns_ingen    = NS(world, "http://drobilla.net/ns/ingen#")
    ns_mod      = NS(world, "http://moddevices.com/ns/mod#")
    ns_modpedal = NS(world, "http://moddevices.com/ns/modpedal#")

    # check if the plugin is a pedalboard
    def fill_in_type(node):
        return node.as_string()
    plugin_types = [i for i in LILV_FOREACH(plugin.get_value(ns_rdf.type_), fill_in_type)]

    if "http://moddevices.com/ns/modpedal#Pedalboard" not in plugin_types:
        raise Exception('get_pedalboard_info(%s) - plugin has no mod:Pedalboard type'.format(bundle))

    # let's get all the info now
    ingenarcs   = []
    ingenblocks = []

    info = {
        'name'  : plugin.get_name().as_string(),
        'uri'   : plugin.get_uri().as_string(),
        'author': plugin.get_author_name().as_string() or "", # Might be empty
        'hardware': {
            # we save this info later
            'audio': {
                'ins' : 0,
                'outs': 0
             },
            'cv': {
                'ins' : 0,
                'outs': 0
             },
            'midi': {
                'ins' : 0,
                'outs': 0
             }
        },
        'size': {
            'width' : plugin.get_value(ns_modpedal.width).get_first().as_int(),
            'height': plugin.get_value(ns_modpedal.height).get_first().as_int(),
        },
        'screenshot' : os.path.basename(plugin.get_value(ns_modpedal.screenshot).get_first().as_string() or ""),
        'thumbnail'  : os.path.basename(plugin.get_value(ns_modpedal.thumbnail).get_first().as_string() or ""),
        'connections': [], # we save this info later
        'plugins'    : []  # we save this info later
    }

    # connections
    arcs = plugin.get_value(ns_ingen.arc)
    it = arcs.begin()
    while not arcs.is_end(it):
        arc = arcs.get(it)
        it  = arcs.next(it)

        if arc.me is None:
            continue

        head = lilv.lilv_world_get(world.me, arc.me, ns_ingen.head.me, None)
        tail = lilv.lilv_world_get(world.me, arc.me, ns_ingen.tail.me, None)

        if head is None or tail is None:
            continue

        ingenarcs.append({
            "source": lilv.lilv_uri_to_path(lilv.lilv_node_as_string(tail)).replace(bundle,"",1),
            "target": lilv.lilv_uri_to_path(lilv.lilv_node_as_string(head)).replace(bundle,"",1)
        })

    # hardware ports
    handled_port_uris = []
    ports = plugin.get_value(ns_lv2core.port)
    it = ports.begin()
    while not ports.is_end(it):
        port = ports.get(it)
        it   = ports.next(it)

        if port.me is None:
            continue

        # check if we already handled this port
        port_uri = port.as_uri()
        if port_uri in handled_port_uris:
            continue
        if port_uri.endswith("/control_in") or port_uri.endswith("/control_out"):
            continue
        handled_port_uris.append(port_uri)

        # get types
        port_types = lilv.lilv_world_find_nodes(world.me, port.me, ns_rdf.type_.me, None)

        if port_types is None:
            continue

        portDir  = "" # input or output
        portType = "" # atom, audio or cv

        it2 = lilv.lilv_nodes_begin(port_types)
        while not lilv.lilv_nodes_is_end(port_types, it2):
            port_type = lilv.lilv_nodes_get(port_types, it2)
            it2 = lilv.lilv_nodes_next(port_types, it2)

            if port_type is None:
                continue

            port_type_uri = lilv.lilv_node_as_uri(port_type)

            if port_type_uri == "http://lv2plug.in/ns/lv2core#InputPort":
                portDir = "input"
            elif port_type_uri == "http://lv2plug.in/ns/lv2core#OutputPort":
                portDir = "output"
            elif port_type_uri == "http://lv2plug.in/ns/lv2core#AudioPort":
                portType = "audio"
            elif port_type_uri == "http://lv2plug.in/ns/lv2core#CVPort":
                portType = "cv"
            elif port_type_uri == "http://lv2plug.in/ns/ext/atom#AtomPort":
                portType = "atom"

        if not (portDir or portType):
            continue

        if portType == "audio":
            if portDir == "input":
                info['hardware']['audio']['ins'] += 1
            else:
                info['hardware']['audio']['outs'] += 1

        elif portType == "atom":
            if portDir == "input":
                info['hardware']['midi']['ins'] += 1
            else:
                info['hardware']['midi']['outs'] += 1

        elif portType == "cv":
            if portDir == "input":
                info['hardware']['cv']['ins'] += 1
            else:
                info['hardware']['cv']['outs'] += 1

    # plugins
    blocks = plugin.get_value(ns_ingen.block)
    it = blocks.begin()
    while not blocks.is_end(it):
        block = blocks.get(it)
        it    = blocks.next(it)

        if block.me is None:
            continue

        protouri1 = lilv.lilv_world_get(world.me, block.me, ns_lv2core.prototype.me, None)
        protouri2 = lilv.lilv_world_get(world.me, block.me, ns_ingen.prototype.me, None)

        if protouri1 is not None:
            proto = protouri1
        elif protouri2 is not None:
            proto = protouri2
        else:
            continue

        instance = lilv.lilv_uri_to_path(lilv.lilv_node_as_string(block.me)).replace(bundle,"",1)
        uri      = lilv.lilv_node_as_uri(proto)

        enabled  = lilv.lilv_world_get(world.me, block.me, ns_ingen.enabled.me, None)
        builder  = lilv.lilv_world_get(world.me, block.me, ns_mod.builderVersion.me, None)
        release  = lilv.lilv_world_get(world.me, block.me, ns_mod.releaseNumber.me, None)
        minorver = lilv.lilv_world_get(world.me, block.me, ns_lv2core.minorVersion.me, None)
        microver = lilv.lilv_world_get(world.me, block.me, ns_lv2core.microVersion.me, None)

        ingenblocks.append({
            "instance": instance,
            "uri"     : uri,
            "x"       : lilv.lilv_node_as_float(lilv.lilv_world_get(world.me, block.me, ns_ingen.canvasX.me, None)),
            "y"       : lilv.lilv_node_as_float(lilv.lilv_world_get(world.me, block.me, ns_ingen.canvasY.me, None)),
            "enabled" : lilv.lilv_node_as_bool(enabled) if enabled is not None else False,
            "builder" : lilv.lilv_node_as_int(builder) if builder else 0,
            "release" : lilv.lilv_node_as_int(release) if release else 0,
            "minorVersion": lilv.lilv_node_as_int(minorver) if minorver else 0,
            "microVersion": lilv.lilv_node_as_int(microver) if microver else 0,
        })

    info['connections'] = ingenarcs
    info['plugins']     = ingenblocks

    return info

# ------------------------------------------------------------------------------------------------------------
# get_pedalboard_name

# Faster version of get_pedalboard_info when we just need to know the pedalboard name
# @a bundle is a string, consisting of a directory in the filesystem (absolute pathname).
def get_pedalboard_name(bundle):
    # lilv wants the last character as the separator
    bundle = os.path.abspath(bundle)
    if not bundle.endswith(os.sep):
        bundle += os.sep

    # Create our own unique lilv world
    # We'll load a single bundle and get all plugins from it
    world = lilv.World()

    # this is needed when loading specific bundles instead of load_all
    # (these functions are not exposed via World yet)
    lilv.lilv_world_load_specifications(world.me)
    lilv.lilv_world_load_plugin_classes(world.me)

    # convert bundle string into a lilv node
    bundlenode = lilv.lilv_new_file_uri(world.me, None, bundle)

    # load the bundle
    world.load_bundle(bundlenode)

    # free bundlenode, no longer needed
    lilv.lilv_node_free(bundlenode)

    # get all plugins in the bundle
    plugins = world.get_all_plugins()

    # make sure the bundle includes 1 and only 1 plugin (the pedalboard)
    if plugins.size() != 1:
        raise Exception('get_pedalboard_info(%s) - bundle has 0 or > 1 plugin'.format(bundle))

    # no indexing in python-lilv yet, just get the first item
    plugin = None
    for p in plugins:
        plugin = p
        break

    if plugin is None:
        raise Exception('get_pedalboard_info(%s) - failed to get plugin, you are using an old lilv!'.format(bundle))

    # define the needed stuff
    ns_rdf = NS(world, lilv.LILV_NS_RDF)

    # check if the plugin is a pedalboard
    def fill_in_type(node):
        return node.as_string()
    plugin_types = [i for i in LILV_FOREACH(plugin.get_value(ns_rdf.type_), fill_in_type)]

    if "http://moddevices.com/ns/modpedal#Pedalboard" not in plugin_types:
        raise Exception('get_pedalboard_info(%s) - plugin has no mod:Pedalboard type'.format(bundle))

    return plugin.get_name().as_string()

# ------------------------------------------------------------------------------------------------------------
# plugin_has_modgui

# Check if a plugin has modgui
def plugin_has_modgui(world, plugin):
    # define the needed stuff
    ns_modgui = NS(world, "http://moddevices.com/ns/modgui#")

    # --------------------------------------------------------------------------------------------------------
    # get the proper modgui

    modguigui = None

    nodes = plugin.get_value(ns_modgui.gui)
    it    = nodes.begin()
    while not nodes.is_end(it):
        mgui = nodes.get(it)
        it   = nodes.next(it)
        if mgui.me is None:
            continue
        resdir = world.find_nodes(mgui.me, ns_modgui.resourcesDirectory.me, None).get_first()
        if resdir.me is None:
            continue
        modguigui = mgui
        if os.path.expanduser("~") in lilv.lilv_uri_to_path(resdir.as_string()):
            # found a modgui in the home dir, stop here and use it
            break

    del nodes, it

    # --------------------------------------------------------------------------------------------------------
    # check selected modgui

    if modguigui is None or modguigui.me is None:
        return False

    # resourcesDirectory *must* be present
    modgui_resdir = world.find_nodes(modguigui.me, ns_modgui.resourcesDirectory.me, None).get_first()

    if modgui_resdir.me is None:
        return False

    return os.path.exists(lilv.lilv_uri_to_path(modgui_resdir.as_string()))

# ------------------------------------------------------------------------------------------------------------
# get_plugin_info

# Get info from a lilv plugin
# This is used in get_plugins_info below and MOD-SDK
def get_plugin_info(world, plugin, useAbsolutePath = True):
    # define the needed stuff
    ns_doap    = NS(world, lilv.LILV_NS_DOAP)
    ns_foaf    = NS(world, lilv.LILV_NS_FOAF)
    ns_rdf     = NS(world, lilv.LILV_NS_RDF)
    ns_rdfs    = NS(world, lilv.LILV_NS_RDFS)
    ns_lv2core = NS(world, lilv.LILV_NS_LV2)
    ns_atom    = NS(world, "http://lv2plug.in/ns/ext/atom#")
    ns_midi    = NS(world, "http://lv2plug.in/ns/ext/midi#")
    ns_morph   = NS(world, "http://lv2plug.in/ns/ext/morph#")
    ns_pprops  = NS(world, "http://lv2plug.in/ns/ext/port-props#")
    ns_pset    = NS(world, "http://lv2plug.in/ns/ext/presets#")
    ns_units   = NS(world, "http://lv2plug.in/ns/extensions/units#")
    ns_mod     = NS(world, "http://moddevices.com/ns/mod#")
    ns_modgui  = NS(world, "http://moddevices.com/ns/modgui#")

    bundleuri = plugin.get_bundle_uri().as_string()
    bundle    = lilv.lilv_uri_to_path(bundleuri)

    errors   = []
    warnings = []

    # --------------------------------------------------------------------------------------------------------
    # uri

    uri = plugin.get_uri().as_string() or ""

    if not uri:
        errors.append("plugin uri is missing or invalid")
    elif uri.startswith("file:"):
        errors.append("plugin uri is local, and thus not suitable for redistribution")
    #elif not (uri.startswith("http:") or uri.startswith("https:")):
        #warnings.append("plugin uri is not a real url")

    # --------------------------------------------------------------------------------------------------------
    # name

    name = plugin.get_name().as_string() or ""

    if not name:
        errors.append("plugin name is missing")

    # --------------------------------------------------------------------------------------------------------
    # binary

    binary = lilv.lilv_uri_to_path(plugin.get_library_uri().as_string() or "")

    if not binary:
        errors.append("plugin binary is missing")
    elif not useAbsolutePath:
        binary = binary.replace(bundle,"",1)

    # --------------------------------------------------------------------------------------------------------
    # license

    license = plugin.get_value(ns_doap.license).get_first().as_string() or ""

    if not license:
        prj = plugin.get_value(ns_lv2core.project).get_first()
        if prj.me is not None:
            licsnode = lilv.lilv_world_get(world.me, prj.me, ns_doap.license.me, None)
            if licsnode is not None:
                license = lilv.lilv_node_as_string(licsnode)
            del licsnode
        del prj

    if not license:
        errors.append("plugin license is missing")

    elif license.startswith(bundleuri):
        license = license.replace(bundleuri,"",1)
        warnings.append("plugin license entry is a local path instead of a string")

    # --------------------------------------------------------------------------------------------------------
    # comment

    comment = (plugin.get_value(ns_rdfs.comment).get_first().as_string() or "").strip()

    # sneaky empty comments!
    if len(comment) > 0 and comment == len(comment) * comment[0]:
        comment = ""

    if not comment:
        errors.append("plugin comment is missing")

    # --------------------------------------------------------------------------------------------------------
    # version

    microver = plugin.get_value(ns_lv2core.microVersion).get_first()
    minorver = plugin.get_value(ns_lv2core.minorVersion).get_first()

    if microver.me is None and minorver.me is None:
        errors.append("plugin is missing version information")
        minorVersion = 0
        microVersion = 0

    else:
        if minorver.me is None:
            errors.append("plugin is missing minorVersion")
            minorVersion = 0
        else:
            minorVersion = minorver.as_int()

        if microver.me is None:
            errors.append("plugin is missing microVersion")
            microVersion = 0
        else:
            microVersion = microver.as_int()

    del minorver
    del microver

    version = "%d.%d" % (minorVersion, microVersion)

    # 0.x is experimental
    if minorVersion == 0:
        stability = "experimental"

    # odd x.2 or 2.x is testing/development
    elif minorVersion % 2 != 0 or microVersion % 2 != 0:
        stability = "testing"

    # otherwise it's stable
    else:
        stability = "stable"

    # --------------------------------------------------------------------------------------------------------
    # author

    author = {
        'name'    : plugin.get_author_name().as_string() or "",
        'homepage': plugin.get_author_homepage().as_string() or "",
        'email'   : plugin.get_author_email().as_string() or "",
    }

    if not author['name']:
        errors.append("plugin author name is missing")

    if not author['homepage']:
        prj = plugin.get_value(ns_lv2core.project).get_first()
        if prj.me is not None:
            maintainer = lilv.lilv_world_get(world.me, prj.me, ns_doap.maintainer.me, None)
            if maintainer is not None:
                homepage = lilv.lilv_world_get(world.me, maintainer, ns_foaf.homepage.me, None)
                if homepage is not None:
                    author['homepage'] = lilv.lilv_node_as_string(homepage)
                del homepage
            del maintainer
        del prj

    if not author['homepage']:
        warnings.append("plugin author homepage is missing")

    if not author['email']:
        pass
    elif author['email'].startswith(bundleuri):
        author['email'] = author['email'].replace(bundleuri,"",1)
        warnings.append("plugin author email entry is missing 'mailto:' prefix")
    elif author['email'].startswith("mailto:"):
        author['email'] = author['email'].replace("mailto:","",1)

    # --------------------------------------------------------------------------------------------------------
    # brand

    brand = plugin.get_value(ns_mod.brand).get_first().as_string() or ""

    if not brand:
        brand = author['name'].split(" - ",1)[0].split(" ",1)[0]
        brand = brand.rstrip(",").rstrip(";")
        if len(brand) > 11:
            brand = brand[:11]
        warnings.append("plugin brand is missing")

    elif len(brand) > 11:
        brand = brand[:11]
        errors.append("plugin brand has more than 11 characters")

    # --------------------------------------------------------------------------------------------------------
    # label

    label = plugin.get_value(ns_mod.label).get_first().as_string() or ""

    if not label:
        if len(name) <= 16:
            label = name
        else:
            labels = name.split(" - ",1)[0].split(" ")
            if labels[0].lower() in bundle.lower() and len(labels) > 1 and not labels[1].startswith(("(","[")):
                label = labels[1]
            else:
                label = labels[0]

            if len(label) > 16:
                label = label[:16]

            warnings.append("plugin label is missing")
            del labels

    elif len(label) > 16:
        label = label[:16]
        errors.append("plugin label has more than 16 characters")

    # --------------------------------------------------------------------------------------------------------
    # bundles

    bundles = []

    if useAbsolutePath:
        bnodes = lilv.lilv_plugin_get_data_uris(plugin.me)

        it = lilv.lilv_nodes_begin(bnodes)
        while not lilv.lilv_nodes_is_end(bnodes, it):
            bnode = lilv.lilv_nodes_get(bnodes, it)
            it    = lilv.lilv_nodes_next(bnodes, it)

            if bnode is None:
                continue
            if not lilv.lilv_node_is_uri(bnode):
                continue

            bpath = os.path.abspath(os.path.dirname(lilv.lilv_uri_to_path(lilv.lilv_node_as_uri(bnode))))

            if not bpath.endswith(os.sep):
                bpath += os.sep

            if bpath not in bundles:
                bundles.append(bpath)

        if bundle not in bundles:
            bundles.append(bundle)

        del bnodes, it

    # --------------------------------------------------------------------------------------------------------
    # get the proper modgui

    modguigui = None

    nodes = plugin.get_value(ns_modgui.gui)
    it    = nodes.begin()
    while not nodes.is_end(it):
        mgui = nodes.get(it)
        it   = nodes.next(it)
        if mgui.me is None:
            continue
        resdir = world.find_nodes(mgui.me, ns_modgui.resourcesDirectory.me, None).get_first()
        if resdir.me is None:
            continue
        modguigui = mgui
        if not useAbsolutePath:
            # special build, use first modgui found
            break
        if os.path.expanduser("~") in lilv.lilv_uri_to_path(resdir.as_string()):
            # found a modgui in the home dir, stop here and use it
            break

    del nodes, it

    # --------------------------------------------------------------------------------------------------------
    # gui

    gui = {}

    if modguigui is None or modguigui.me is None:
        warnings.append("no modgui available")

    else:
        # resourcesDirectory *must* be present
        modgui_resdir = world.find_nodes(modguigui.me, ns_modgui.resourcesDirectory.me, None).get_first()

        if modgui_resdir.me is None:
            errors.append("modgui has no resourcesDirectory data")

        else:
            if useAbsolutePath:
                gui['resourcesDirectory'] = lilv.lilv_uri_to_path(modgui_resdir.as_string())

                # check if modgui is defined in a separate file
                gui['usingSeeAlso'] = os.path.exists(os.path.join(bundle, "modgui.ttl"))

                # check if the modgui definition is on its own file and in the user dir
                gui['modificableInPlace'] = bool((bundle not in gui['resourcesDirectory'] or gui['usingSeeAlso']) and
                                                os.path.expanduser("~") in gui['resourcesDirectory'])
            else:
                gui['resourcesDirectory'] = modgui_resdir.as_string().replace(bundleuri,"",1)

            # icon and settings templates
            modgui_icon  = world.find_nodes(modguigui.me, ns_modgui.iconTemplate    .me, None).get_first()
            modgui_setts = world.find_nodes(modguigui.me, ns_modgui.settingsTemplate.me, None).get_first()

            if modgui_icon.me is None:
                errors.append("modgui has no iconTemplate data")
            else:
                iconFile = lilv.lilv_uri_to_path(modgui_icon.as_string())
                if os.path.exists(iconFile):
                    gui['iconTemplate'] = iconFile if useAbsolutePath else iconFile.replace(bundle,"",1)
                else:
                    errors.append("modgui iconTemplate file is missing")
                del iconFile

            if modgui_setts.me is not None:
                settingsFile = lilv.lilv_uri_to_path(modgui_setts.as_string())
                if os.path.exists(settingsFile):
                    gui['settingsTemplate'] = settingsFile if useAbsolutePath else settingsFile.replace(bundle,"",1)
                else:
                    errors.append("modgui settingsTemplate file is missing")
                del settingsFile

            # javascript and stylesheet files
            modgui_script = world.find_nodes(modguigui.me, ns_modgui.javascript.me, None).get_first()
            modgui_style  = world.find_nodes(modguigui.me, ns_modgui.stylesheet.me, None).get_first()

            if modgui_script.me is not None:
                javascriptFile = lilv.lilv_uri_to_path(modgui_script.as_string())
                if os.path.exists(javascriptFile):
                    gui['javascript'] = javascriptFile if useAbsolutePath else javascriptFile.replace(bundle,"",1)
                else:
                    errors.append("modgui javascript file is missing")
                del javascriptFile

            if modgui_style.me is None:
                errors.append("modgui has no stylesheet data")
            else:
                stylesheetFile = lilv.lilv_uri_to_path(modgui_style.as_string())
                if os.path.exists(stylesheetFile):
                    gui['stylesheet'] = stylesheetFile if useAbsolutePath else stylesheetFile.replace(bundle,"",1)
                else:
                    errors.append("modgui stylesheet file is missing")
                del stylesheetFile

            # template data for backwards compatibility
            # FIXME remove later once we got rid of all templateData files
            modgui_templ = world.find_nodes(modguigui.me, ns_modgui.templateData.me, None).get_first()

            if modgui_templ.me is not None:
                warnings.append("modgui is using old deprecated templateData")
                templFile = lilv.lilv_uri_to_path(modgui_templ.as_string())
                if os.path.exists(templFile):
                    with open(templFile, 'r') as fd:
                        try:
                            data = json.loads(fd.read())
                        except:
                            data = {}
                        keys = list(data.keys())

                        if 'author' in keys:
                            gui['brand'] = data['author']
                        if 'label' in keys:
                            gui['label'] = data['label']
                        if 'color' in keys:
                            gui['color'] = data['color']
                        if 'knob' in keys:
                            gui['knob'] = data['knob']
                        if 'controls' in keys:
                            index = 0
                            ports = []
                            for ctrl in data['controls']:
                                ports.append({
                                    'index' : index,
                                    'name'  : ctrl['name'],
                                    'symbol': ctrl['symbol'],
                                })
                                index += 1
                            gui['ports'] = ports
                del templFile

            # screenshot and thumbnail
            modgui_scrn  = world.find_nodes(modguigui.me, ns_modgui.screenshot.me, None).get_first()
            modgui_thumb = world.find_nodes(modguigui.me, ns_modgui.thumbnail .me, None).get_first()

            if modgui_scrn.me is not None:
                gui['screenshot'] = lilv.lilv_uri_to_path(modgui_scrn.as_string())
                if not os.path.exists(gui['screenshot']):
                    errors.append("modgui screenshot file is missing")
                if not useAbsolutePath:
                    gui['screenshot'] = gui['screenshot'].replace(bundle,"",1)
            else:
                errors.append("modgui has no screnshot data")

            if modgui_thumb.me is not None:
                gui['thumbnail'] = lilv.lilv_uri_to_path(modgui_thumb.as_string())
                if not os.path.exists(gui['thumbnail']):
                    errors.append("modgui thumbnail file is missing")
                if not useAbsolutePath:
                    gui['thumbnail'] = gui['thumbnail'].replace(bundle,"",1)
            else:
                errors.append("modgui has no thumbnail data")

            # extra stuff, all optional
            modgui_brand = world.find_nodes(modguigui.me, ns_modgui.brand.me, None).get_first()
            modgui_label = world.find_nodes(modguigui.me, ns_modgui.label.me, None).get_first()
            modgui_model = world.find_nodes(modguigui.me, ns_modgui.model.me, None).get_first()
            modgui_panel = world.find_nodes(modguigui.me, ns_modgui.panel.me, None).get_first()
            modgui_color = world.find_nodes(modguigui.me, ns_modgui.color.me, None).get_first()
            modgui_knob  = world.find_nodes(modguigui.me, ns_modgui.knob .me, None).get_first()

            if modgui_brand.me is not None:
                gui['brand'] = modgui_brand.as_string()
            if modgui_label.me is not None:
                gui['label'] = modgui_label.as_string()
            if modgui_model.me is not None:
                gui['model'] = modgui_model.as_string()
            if modgui_panel.me is not None:
                gui['panel'] = modgui_panel.as_string()
            if modgui_color.me is not None:
                gui['color'] = modgui_color.as_string()
            if modgui_knob.me is not None:
                gui['knob'] = modgui_knob.as_string()

            # ports
            errpr = False
            sybls = []
            ports = []
            nodes = world.find_nodes(modguigui.me, ns_modgui.port.me, None)
            it    = lilv.lilv_nodes_begin(nodes.me)
            while not lilv.lilv_nodes_is_end(nodes.me, it):
                port = lilv.lilv_nodes_get(nodes.me, it)
                it   = lilv.lilv_nodes_next(nodes.me, it)
                if port is None:
                    break
                port_indx = world.find_nodes(port, ns_lv2core.index .me, None).get_first()
                port_symb = world.find_nodes(port, ns_lv2core.symbol.me, None).get_first()
                port_name = world.find_nodes(port, ns_lv2core.name  .me, None).get_first()

                if None in (port_indx.me, port_name.me, port_symb.me):
                    if not errpr:
                        errors.append("modgui has some invalid port data")
                        errpr = True
                    continue

                port_indx = port_indx.as_int()
                port_symb = port_symb.as_string()
                port_name = port_name.as_string()

                ports.append({
                    'index' : port_indx,
                    'symbol': port_symb,
                    'name'  : port_name,
                })

                if port_symb not in sybls:
                    sybls.append(port_symb)
                elif not errpr:
                    errors.append("modgui has some duplicated port symbols")
                    errpr = True

            # sort ports
            if len(ports) > 0:
                ports2 = {}

                for port in ports:
                    ports2[port['index']] = port
                gui['ports'] = [ports2[i] for i in ports2]

                del ports2

            # cleanup
            del ports, nodes, it

    # --------------------------------------------------------------------------------------------------------
    # ports

    index = 0
    ports = {
        'audio'  : { 'input': [], 'output': [] },
        'control': { 'input': [], 'output': [] },
        'midi'   : { 'input': [], 'output': [] }
    }

    portsymbols = []
    portnames   = []

    # function for filling port info
    def fill_port_info(port):
        # base data
        portname = lilv.lilv_node_as_string(port.get_name()) or ""

        if not portname:
            portname = "_%i" % index
            errors.append("port with index %i has no name" % index)

        portsymbol = lilv.lilv_node_as_string(port.get_symbol()) or ""

        if not portsymbol:
            portsymbol = "_%i" % index
            errors.append("port with index %i has no symbol" % index)

        # check for duplicate names
        if portname in portsymbols:
            warnings.append("port name '%s' is not unique" % portname)
        else:
            portnames.append(portname)

        # check for duplicate symbols
        if portsymbol in portsymbols:
            errors.append("port symbol '%s' is not unique" % portsymbol)
        else:
            portsymbols.append(portsymbol)

        # short name
        psname = lilv.lilv_nodes_get_first(port.get_value(ns_lv2core.shortName.me))

        if psname is not None:
            psname = lilv.lilv_node_as_string(psname) or ""

        if not psname:
            psname = get_short_port_name(portname)
            if len(psname) > 16:
                warnings.append("port '%s' name is too big, reduce the name size or provide a shortName" % portname)

        elif len(psname) > 16:
            psname = psname[:16]
            errors.append("port '%s' short name has more than 16 characters" % portname)

        # check for old style shortName
        if port.get_value(ns_lv2core.shortname.me) is not None:
            errors.append("port '%s' short name is using old style 'shortname' instead of 'shortName'" % portname)

        # port types
        types = [typ.rsplit("#",1)[-1].replace("Port","",1) for typ in get_port_data(port, ns_rdf.type_)]

        if "Atom" in types \
            and port.supports_event(ns_midi.MidiEvent.me) \
            and lilv.Nodes(port.get_value(ns_atom.bufferType.me)).get_first() == ns_atom.Sequence:
                types.append("MIDI")

        #if "Morph" in types:
            #morphtyp = lilv.lilv_nodes_get_first(port.get_value(ns_morph.supportsType.me))
            #if morphtyp is not None:
                #morphtyp = lilv.lilv_node_as_uri(morphtyp)
                #if morphtyp:
                    #types.append(morphtyp.rsplit("#",1)[-1].replace("Port","",1))

        # port comment
        pcomment = (get_port_data(port, ns_rdfs.comment) or [""])[0]

        # port designation
        designation = (get_port_data(port, ns_lv2core.designation) or [""])[0]

        # port rangeSteps
        rangeSteps = (get_port_data(port, ns_mod.rangeSteps) or get_port_data(port, ns_pprops.rangeSteps) or [None])[0]

        # port properties
        properties = [typ.rsplit("#",1)[-1] for typ in get_port_data(port, ns_lv2core.portProperty)]

        # data
        ranges      = {}
        scalepoints = []

        # unit block
        ulabel  = ""
        urender = ""
        usymbol = ""

        # control and cv must contain ranges, might contain scale points
        if "Control" in types or "CV" in types:
            isInteger = "integer" in properties

            if isInteger and "CV" in types:
                errors.append("port '%s' has integer property and CV type" % portname)

            xdefault = lilv.lilv_nodes_get_first(port.get_value(ns_mod.default.me)) or \
                       lilv.lilv_nodes_get_first(port.get_value(ns_lv2core.default.me))
            xminimum = lilv.lilv_nodes_get_first(port.get_value(ns_mod.minimum.me)) or \
                       lilv.lilv_nodes_get_first(port.get_value(ns_lv2core.minimum.me))
            xmaximum = lilv.lilv_nodes_get_first(port.get_value(ns_mod.maximum.me)) or \
                       lilv.lilv_nodes_get_first(port.get_value(ns_lv2core.maximum.me))

            if xminimum is not None and xmaximum is not None:
                if isInteger:
                    if is_integer(lilv.lilv_node_as_string(xminimum)):
                        ranges['minimum'] = lilv.lilv_node_as_int(xminimum)
                    else:
                        ranges['minimum'] = lilv.lilv_node_as_float(xminimum)
                        if fmod(ranges['minimum'], 1.0) == 0.0:
                            warnings.append("port '%s' has integer property but minimum value is float" % portname)
                        else:
                            errors.append("port '%s' has integer property but minimum value has non-zero decimals" % portname)
                        ranges['minimum'] = int(ranges['minimum'])

                    if is_integer(lilv.lilv_node_as_string(xmaximum)):
                        ranges['maximum'] = lilv.lilv_node_as_int(xmaximum)
                    else:
                        ranges['maximum'] = lilv.lilv_node_as_float(xmaximum)
                        if fmod(ranges['maximum'], 1.0) == 0.0:
                            warnings.append("port '%s' has integer property but maximum value is float" % portname)
                        else:
                            errors.append("port '%s' has integer property but maximum value has non-zero decimals" % portname)
                        ranges['maximum'] = int(ranges['maximum'])

                else:
                    ranges['minimum'] = lilv.lilv_node_as_float(xminimum)
                    ranges['maximum'] = lilv.lilv_node_as_float(xmaximum)

                    if is_integer(lilv.lilv_node_as_string(xminimum)):
                        warnings.append("port '%s' minimum value is an integer" % portname)

                    if is_integer(lilv.lilv_node_as_string(xmaximum)):
                        warnings.append("port '%s' maximum value is an integer" % portname)

                if ranges['minimum'] >= ranges['maximum']:
                    ranges['maximum'] = ranges['minimum'] + (1 if isInteger else 0.1)
                    errors.append("port '%s' minimum value is equal or higher than its maximum" % portname)

                if xdefault is not None:
                    if isInteger:
                        if is_integer(lilv.lilv_node_as_string(xdefault)):
                            ranges['default'] = lilv.lilv_node_as_int(xdefault)
                        else:
                            ranges['default'] = lilv.lilv_node_as_float(xdefault)
                            if fmod(ranges['default'], 1.0) == 0.0:
                                warnings.append("port '%s' has integer property but default value is float" % portname)
                            else:
                                errors.append("port '%s' has integer property but default value has non-zero decimals" % portname)
                            ranges['default'] = int(ranges['default'])
                    else:
                        ranges['default'] = lilv.lilv_node_as_float(xdefault)

                        if is_integer(lilv.lilv_node_as_string(xdefault)):
                            warnings.append("port '%s' default value is an integer" % portname)

                    testmin = ranges['minimum']
                    testmax = ranges['maximum']

                    if "sampleRate" in properties:
                        testmin *= 48000
                        testmax *= 48000

                    if not (testmin <= ranges['default'] <= testmax):
                        ranges['default'] = ranges['minimum']
                        errors.append("port '%s' default value is out of bounds" % portname)

                else:
                    ranges['default'] = ranges['minimum']

                    if "Input" in types:
                        errors.append("port '%s' is missing default value" % portname)

            else:
                if isInteger:
                    ranges['minimum'] = 0
                    ranges['maximum'] = 1
                    ranges['default'] = 0
                else:
                    ranges['minimum'] = -1.0 if "CV" in types else 0.0
                    ranges['maximum'] = 1.0
                    ranges['default'] = 0.0

                if "CV" not in types and designation != "http://lv2plug.in/ns/lv2core#latency":
                    errors.append("port '%s' is missing value ranges" % portname)

            nodes = port.get_scale_points()

            if nodes is not None:
                scalepoints_unsorted = []

                it = lilv.lilv_scale_points_begin(nodes)
                while not lilv.lilv_scale_points_is_end(nodes, it):
                    sp = lilv.lilv_scale_points_get(nodes, it)
                    it = lilv.lilv_scale_points_next(nodes, it)

                    if sp is None:
                        continue

                    label = lilv.lilv_scale_point_get_label(sp)
                    value = lilv.lilv_scale_point_get_value(sp)

                    if label is None:
                        errors.append("a port scalepoint is missing its label")
                        continue

                    label = lilv.lilv_node_as_string(label) or ""

                    if not label:
                        errors.append("a port scalepoint is missing its label")
                        continue

                    if value is None:
                        errors.append("port scalepoint '%s' is missing its value" % label)
                        continue

                    if isInteger:
                        if is_integer(lilv.lilv_node_as_string(value)):
                            value = lilv.lilv_node_as_int(value)
                        else:
                            value = lilv.lilv_node_as_float(value)
                            if fmod(value, 1.0) == 0.0:
                                warnings.append("port '%s' has integer property but scalepoint '%s' value is float" % (portname, label))
                            else:
                                errors.append("port '%s' has integer property but scalepoint '%s' value has non-zero decimals" % (portname, label))
                            value = int(value)
                    else:
                        if is_integer(lilv.lilv_node_as_string(value)):
                            warnings.append("port '%s' scalepoint '%s' value is an integer" % (portname, label))
                        value = lilv.lilv_node_as_float(value)

                    if ranges['minimum'] <= value <= ranges['maximum']:
                        scalepoints_unsorted.append((value, label))
                    else:
                        errors.append(("port scalepoint '%s' has an out-of-bounds value:\n" % label) +
                                      ("%d < %d < %d" if isInteger else "%f < %f < %f") % (ranges['minimum'], value, ranges['maximum']))

                if len(scalepoints_unsorted) != 0:
                    unsorted = dict(s for s in scalepoints_unsorted)
                    values   = list(v for v, l in scalepoints_unsorted)
                    values.sort()
                    scalepoints = list({ 'value': v, 'label': unsorted[v] } for v in values)
                    del unsorted, values

                del scalepoints_unsorted

            if "enumeration" in properties and len(scalepoints) <= 1:
                errors.append("port '%s' wants to use enumeration but doesn't have enough values" % portname)
                properties.remove("enumeration")

        # control ports might contain unit
        if "Control" in types:
            # unit
            uunit = lilv.lilv_nodes_get_first(port.get_value(ns_units.unit.me))

            if uunit is not None:
                uuri = lilv.lilv_node_as_uri(uunit)

                # using pre-existing lv2 unit
                if uuri is not None and uuri.startswith("http://lv2plug.in/ns/"):
                    uuri  = uuri.replace("http://lv2plug.in/ns/extensions/units#","",1)
                    alnum = uuri.isalnum()

                    if not alnum:
                        errors.append("port '%s' has wrong lv2 unit uri" % portname)
                        uuri = uuri.rsplit("#",1)[-1].rsplit("/",1)[-1]

                    ulabel, urender, usymbol = get_port_unit(uuri)

                    if alnum and not (ulabel and urender and usymbol):
                        errors.append("port '%s' has unknown lv2 unit (our bug?, data is '%s', '%s', '%s')" % (portname,
                                                                                                               ulabel,
                                                                                                               urender,
                                                                                                               usymbol))

                # using custom unit
                else:
                    xlabel  = world.find_nodes(uunit, ns_rdfs  .label.me, None).get_first()
                    xrender = world.find_nodes(uunit, ns_units.render.me, None).get_first()
                    xsymbol = world.find_nodes(uunit, ns_units.symbol.me, None).get_first()

                    if xlabel.me is not None:
                        ulabel = xlabel.as_string()
                    else:
                        errors.append("port '%s' has custom unit with no label" % portname)

                    if xrender.me is not None:
                        urender = xrender.as_string()
                    else:
                        errors.append("port '%s' has custom unit with no render" % portname)

                    if xsymbol.me is not None:
                        usymbol = xsymbol.as_string()
                    else:
                        errors.append("port '%s' has custom unit with no symbol" % portname)

        return (types, {
            'name'   : portname,
            'symbol' : portsymbol,
            'ranges' : ranges,
            'units'  : {
                'label' : ulabel,
                'render': urender,
                'symbol': usymbol,
            } if "Control" in types and ulabel and urender and usymbol else {},
            'comment'    : pcomment,
            'designation': designation,
            'properties' : properties,
            'rangeSteps' : rangeSteps,
            'scalePoints': scalepoints,
            'shortName'  : psname,
        })

    for p in (plugin.get_port_by_index(i) for i in range(plugin.get_num_ports())):
        types, info = fill_port_info(p)

        info['index'] = index
        index += 1

        isInput = "Input" in types
        types.remove("Input" if isInput else "Output")

        for typ in [typl.lower() for typl in types]:
            if typ not in ports.keys():
                ports[typ] = { 'input': [], 'output': [] }
            ports[typ]["input" if isInput else "output"].append(info)

    # --------------------------------------------------------------------------------------------------------
    # presets

    def get_preset_data(preset):
        world.load_resource(preset.me)

        uri   = preset.as_string() or ""
        label = world.find_nodes(preset.me, ns_rdfs.label.me, None).get_first().as_string() or ""

        if not uri:
            errors.append("preset with label '%s' has no uri" % (label or "<unknown>"))
        if not label:
            errors.append("preset with uri '%s' has no label" % (uri or "<unknown>"))

        return (uri, label)

    presets = []

    presets_related = plugin.get_related(ns_pset.Preset)
    presets_data    = list(LILV_FOREACH(presets_related, get_preset_data))

    if len(presets_data) != 0:
        unsorted = dict(p for p in presets_data)
        uris     = list(unsorted.keys())
        uris.sort()
        presets  = list({ 'uri': p, 'label': unsorted[p] } for p in uris)
        del unsorted, uris

    del presets_related

    # --------------------------------------------------------------------------------------------------------
    # done

    return {
        'uri' : uri,
        'name': name,

        'binary' : binary,
        'brand'  : brand,
        'label'  : label,
        'license': license,
        'comment': comment,

        'category'    : get_category(plugin.get_value(ns_rdf.type_)),
        'microVersion': microVersion,
        'minorVersion': minorVersion,

        'version'  : version,
        'stability': stability,

        'author' : author,
        'bundles': bundles,
        'gui'    : gui,
        'ports'  : ports,
        'presets': presets,

        'errors'  : errors,
        'warnings': warnings,
    }

# ------------------------------------------------------------------------------------------------------------
# get_plugin_info_helper

# Get info from a simple URI, without the need of your own lilv world
# This is used by get_plugins_info in MOD-SDK
def get_plugin_info_helper(uri):
    world = lilv.World()
    world.load_all()
    plugins = world.get_all_plugins()
    return [get_plugin_info(world, p, False) for p in plugins]

# ------------------------------------------------------------------------------------------------------------
# get_plugins_info

# Get plugin-related info from a list of lv2 bundles
# @a bundles is a list of strings, consisting of directories in the filesystem (absolute pathnames).
def get_plugins_info(bundles):
    # if empty, do nothing
    if len(bundles) == 0:
        raise Exception('get_plugins_info() - no bundles provided')

    # Create our own unique lilv world
    # We'll load the selected bundles and get all plugins from it
    world = lilv.World()

    # this is needed when loading specific bundles instead of load_all
    # (these functions are not exposed via World yet)
    lilv.lilv_world_load_specifications(world.me)
    lilv.lilv_world_load_plugin_classes(world.me)

    # load all bundles
    for bundle in bundles:
        # lilv wants the last character as the separator
        bundle = os.path.abspath(bundle)
        if not bundle.endswith(os.sep):
            bundle += os.sep

        # convert bundle string into a lilv node
        bundlenode = lilv.lilv_new_file_uri(world.me, None, bundle)

        # load the bundle
        world.load_bundle(bundlenode)

        # free bundlenode, no longer needed
        lilv.lilv_node_free(bundlenode)

    # get all plugins available in the selected bundles
    plugins = world.get_all_plugins()

    # make sure the bundles include something
    if plugins.size() == 0:
        raise Exception('get_plugins_info() - selected bundles have no plugins')

    # return all the info
    return [get_plugin_info(world, p, False) for p in plugins]

# ------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    from sys import argv, exit
    from pprint import pprint
    #get_plugins_info(argv[1:])
    #for i in get_plugins_info(argv[1:]): pprint(i)
    #exit(0)
    for i in get_plugins_info(argv[1:]):
        warnings = i['warnings'].copy()

        if 'plugin brand is missing' in warnings:
            i['warnings'].remove('plugin brand is missing')

        if 'plugin label is missing' in warnings:
            i['warnings'].remove('plugin label is missing')

        if 'no modgui available' in warnings:
            i['warnings'].remove('no modgui available')

        for warn in warnings:
            if "has no short name" in warn:
                i['warnings'].remove(warn)

        pprint({
            'uri'     : i['uri'],
            'errors'  : i['errors'],
            'warnings': i['warnings']
        }, width=200)

# ------------------------------------------------------------------------------------------------------------



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.observer.observable_list import ObservableList
from pluginsmanager.observer.update_type import UpdateType

from unittest.mock import MagicMock
from pluginsmanager.observer.autosaver.indexable import Indexable


class Bank(Indexable):
    """
    Bank is a data structure that contains :class:`.Pedalboard`. It's useful
    for group common pedalboards, like "Pedalboards will be used in
    the Sunday show"

    A fast bank overview::

    >>> bank = Bank('RHCP')
    >>> californication = Pedalboard('Californication')

    >>> # Add pedalboard in bank - mode A
    >>> bank.append(californication)
    >>> californication.bank == bank
    True

    >>> bank.pedalboards[0] == californication
    True

    >>> # Add pedalboard in bank - mode B
    >>> bank.pedalboards.append(Pedalboard('Dark Necessities'))
    >>> bank.pedalboards[1].bank == bank
    True

    >>> # If you needs change pedalboards order (swap), use pythonic mode
    >>> bank.pedalboards[1], bank.pedalboards[0] = bank.pedalboards[0], bank.pedalboards[1]
    >>> bank.pedalboards[1] == californication
    True

    >>> # Set pedalboard
    >>> bank.pedalboards[0] = Pedalboard("Can't Stop")
    >>> bank.pedalboards[0].bank == bank
    True

    >>> del bank.pedalboards[0]
    >>> bank.pedalboards[0] == californication # Pedalboard Can't stop rermoved, first is now the californication
    True

    You can also toggle pedalboards into different banks::

    >>> bank1.pedalboards[0], bank2.pedalboards[2] = bank2.pedalboards[0], bank1.pedalboards[2]

    :param string name: Bank name
    """
    def __init__(self, name):
        super(Bank, self).__init__()

        self.name = name
        self.pedalboards = ObservableList()
        self.pedalboards.observer = self._pedalboards_observer

        self.manager = None

        self._observer = MagicMock()

    @property
    def observer(self):
        return self._observer

    @observer.setter
    def observer(self, observer):
        self._observer = observer
        for pedalboard in self.pedalboards:
            pedalboard.observer = observer

    def _pedalboards_observer(self, update_type, pedalboard, index, **kwargs):
        kwargs['index'] = index
        kwargs['origin'] = self

        if update_type == UpdateType.CREATED:
            self._init_pedalboard(pedalboard)

        elif update_type == UpdateType.UPDATED:
            self._init_pedalboard(pedalboard)

            old_pedalboard = kwargs['old']
            if old_pedalboard not in self.pedalboards:
                self._clear_pedalboard(old_pedalboard)

        elif update_type == UpdateType.DELETED:
            self._clear_pedalboard(pedalboard)

        self.observer.on_pedalboard_updated(pedalboard, update_type, **kwargs)

    def _init_pedalboard(self, pedalboard):
        pedalboard.bank = self
        pedalboard.observer = self.observer

    def _clear_pedalboard(self, pedalboard):
        pedalboard.bank = None
        pedalboard.observer = MagicMock()

    @property
    def json(self):
        """
        Get a json decodable representation of this bank

        :return dict: json representation
        """
        return self.__dict__

    @property
    def __dict__(self):
        try:
            index = self.index
        except IndexError:
            index = -1

        return {
            'index': index,
            'name': self.name,
            'pedalboards': [pedalboard.json for pedalboard in self.pedalboards]
        }

    def append(self, pedalboard):
        """
        Add a :class:`.Pedalboard` in this bank

        This works same as::

        >>> bank.pedalboards.append(pedalboard)

        or::

        >>> bank.pedalboards.insert(len(bank.pedalboards), pedalboard)

        :param Pedalboard pedalboard: Pedalboard that will be added
        """
        self.pedalboards.append(pedalboard)

    @property
    def index(self):
        """
        Returns the first occurrence of the bank in your :class:`.PluginsManager`
        """
        if self.manager is None:
            raise IndexError('Bank not contains a manager')

        return self.manager.banks.index(self)

    @property
    def simple_identifier(self):
        return self.name



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from pluginsmanager.model.audio_port import AudioPort


class Output(AudioPort, metaclass=ABCMeta):
    """
    Output is the medium in which the audio processed by the effect is returned.

    Effects usually have a one (mono) or two outputs (stereo L + stereo R). .

    For obtains the outputs::

        >>> my_awesome_effect
        <Lv2Effect object as 'Calf Reverb' active at 0x7fd58d874ba8>
        >>> my_awesome_effect.outputs
        (<Lv2Output object as Out L at 0x7fd58c58a438>, <Lv2Output object as Out R at 0x7fd58c58d550>)

        >>> output = my_awesome_effect.outputs[0]
        >>> output
        <Lv2Output object as Out L at 0x7fd58c58a438>

        >>> symbol = my_awesome_effect.outputs[0].symbol
        >>> symbol
        'output_l'

        >>> my_awesome_effect.outputs[symbol] == output
        True

    For connections between effects, see :meth:`~pluginsmanager.model.pedalboard.Pedalboard.connect()`
    and :meth:`~pluginsmanager.model.pedalboard.Pedalboard.disconnect()` :class:`.Pedalboard` class methods.

    :param Effect effect: Effect that contains the output
    """

    @property
    def index(self):
        """
        :return: Output index in the your effect
        """
        return self.effect.outputs.index(self)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import jack


class JackClient(object):
    """
    Based in: http://jackclient-python.readthedocs.io/en/0.4.2/examples.html#chatty-client

    Through the :class:`jack.JackClient` it is possible to be notified when `x-run`
    occurs and when the Jack server is closed::

        >>> client = JackClient()
        >>> client.xrun_callback = lambda delay: print('x-run', delay)
        >>> client.shutdown_callback = lambda status, reason: print('shutdown: ', status, reason)

    When you don't use anymore, close it::

        >>> client.close()

    :param bool no_start_server: False if starts a new JACK server
                                 True if uses a already started jack (ex: using `jackdump`)
    :param name: Jack client name. Default: `JackClient`
    """
    def __init__(self, no_start_server=True, name=None):
        if name is None:
            name = self.__class__.__name__

        self.client = jack.Client(name=name, no_start_server=no_start_server)

        self.xrun_callback = lambda delay: ...
        self.shutdown_callback = lambda status, reason: ...

        if self.client.status.server_started:
            logging.info('JACK server was started')
        else:
            logging.info('JACK server was already running')

        if self.client.status.name_not_unique:
            logging.info('Unique client name generated {}'.format(self.client.name))

        @self.client.set_xrun_callback
        def xrun(delay):
            self.xrun_callback(delay)

        @self.client.set_shutdown_callback
        def shutdown(status, reason):
            self.shutdown_callback(status, reason)

        self.client.activate()

    @property
    def audio_inputs(self):
        """
        :return: A list of audio input :class:`Ports`.
        """
        return self.client.get_ports(is_audio=True, is_physical=True, is_input=True)

    @property
    def audio_outputs(self):
        """
        :return: A list of audio output :class:`Ports`.
        """
        return self.client.get_ports(is_audio=True, is_physical=True, is_output=True)

    @property
    def midi_inputs(self):
        """
        :return: A list of MIDI input :class:`Ports`.
        """
        return self.client.get_ports(is_midi=True, is_physical=True, is_input=True)

    @property
    def midi_outputs(self):
        """
        :return: A list of MIDI output :class:`Ports`.
        """
        return self.client.get_ports(is_midi=True, is_physical=True, is_output=True)

    def close(self):
        """
        Deactivate and closes the jack client
        """
        self.client.deactivate()
        self.client.close()



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class PairsList(object):
    """
    Receives two lists and generates a result list of pairs of equal elements

    Uses `calculate` method for generate list

    :param similarity_key_function: Function that receives a element and returns your identifier
                                    to do a mapping with elements from another list
    """

    def __init__(self, similarity_key_function):
        self.similarity_key = similarity_key_function

    def calculate(self, list_a, list_b):
        result = PairsListResult()

        hash_elements_a = self._generate_hash(list_a)
        hash_elements_b = self._generate_hash(list_b)

        keys = hash_elements_a.keys() | hash_elements_b.keys()

        # Pairs
        for key in keys:
            if not key in hash_elements_a:
                hash_elements_a[key] = []
            if not key in hash_elements_b:
                hash_elements_b[key] = []

            while hash_elements_a[key] and hash_elements_b[key]:
                element_a = hash_elements_a[key].pop(0)
                element_b = hash_elements_b[key].pop(0)

                result.pairs.append((element_a, element_b))

            result.elements_not_added_a.extend(hash_elements_a[key])
            result.elements_not_added_b.extend(hash_elements_b[key])

        return result

    def _generate_hash(self, lista):
        hash_elements = {}

        for index, element in enumerate(lista):
            key = self.similarity_key(element)

            if key not in hash_elements:
                hash_elements[key] = []

            hash_elements[key].append(element)

        return hash_elements


class PairsListResult(object):
    def __init__(self):
        self.pairs = []
        self.elements_not_added_a = []
        self.elements_not_added_b = []



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta


class SystemPortMixing(object, metaclass=ABCMeta):
    """
    Contains the default implementation of System ports:
    :class:`.SystemInput`, :class:`.SystemOutput`,
    :class:`.SystemMidiInput` and :class:`.SystemMidiInput`
    """

    def __init__(self, *args, **kwargs):
        super(SystemPortMixing, self).__init__(*args, **kwargs)

    def __str__(self):
        return self.symbol

    @property
    def __dict__(self):
        return {
            'symbol': self.symbol,
            'index': self.index,
        }



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pluginsmanager.model.audio_port import AudioPort


class ConnectionError(Exception):
    def __init__(self, message):
        super(ConnectionError, self).__init__(message)
        self.message = message


class Connection(object):
    """
    :class:`~pluginsmanager.model.connection.Connection` represents a connection between two
    distinct effects by your :class:`.AudioPort` (effect :class:`.Output` with effect :class:`.Input`)::

    >>> from pluginsmanager.model.pedalboard import Pedalboard
    >>> californication = Pedalboard('Californication')
    >>> californication.append(driver)
    >>> californication.append(reverb)

    >>> guitar_output = sys_effect.outputs[0]
    >>> driver_input = driver.inputs[0]
    >>> driver_output = driver.outputs[0]
    >>> reverb_input = reverb.inputs[0]
    >>> reverb_output = reverb.outputs[0]
    >>> amp_input = sys_effect.inputs[0]

    >>> # Guitar -> driver -> reverb -> amp
    >>> californication.connections.append(Connection(guitar_output, driver_input))
    >>> californication.connections.append(Connection(driver_output, reverb_input))
    >>> californication.connections.append(Connection(reverb_output, amp_input))

    Another way to use implicitly connections::

    >>> californication.connect(guitar_output, driver_input)
    >>> californication.connect(driver_output, reverb_input)
    >>> californication.connect(reverb_output, amp_input)

    :param Output output_port: Audio output port that will be connected with audio input port
    :param Input input_port: Audio input port that will be connected with audio output port
    """

    def __init__(self, output_port, input_port):
        if not self._valid_instance(output_port)\
        or not self._valid_instance(input_port):
            raise ConnectionError("'{}' only accepts ports that inherits {}".format(self.__class__.__name__, self.ports_class.__name__))

        if output_port.effect == input_port.effect\
        and not output_port.effect.is_possible_connect_itself:
            effect_name = str(input_port.effect)
            raise ConnectionError("The output {} and input {} are from the same effect {}. "
                                  "This effect doesn't accept connections between itself instance."
                                  .format(effect_name, str(output_port), str(input_port)))

        self._output = output_port
        self._input = input_port

    def _valid_instance(self, port):
        return isinstance(port, self.ports_class)

    @property
    def ports_class(self):
        """
        :return class: Port class that this connection only accepts
        """
        return AudioPort

    @property
    def output(self):
        """
        :return Output: Output connection port
        """
        return self._output

    @property
    def input(self):
        """
        :return Output: Input connection port
        """
        return self._input

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
           and self.input == other.input \
           and self.output == other.output

    def __repr__(self):
        return "<{} object as '{}.{} -> {}.{}' at 0x{:x}>".format(
            self.__class__.__name__,
            self.output.effect,
            self.output,
            self.input.effect,
            self.input,
            id(self)
        )

    @property
    def json(self):
        """
        Get a json decodable representation of this effect

        :return dict: json representation
        """
        return self.__dict__

    @property
    def __dict__(self):
        return {
            'output': self.output.json,
            'input': self.input.json,
            'type': 'audio'
        }



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.observer.autosaver.persistence import Persistence


class IndexFile(object):

    def __init__(self, path):
        """
        :param Path path: Index file path
        """
        self.path = path

    def load(self, indexables):
        """
        Load the index file and reorder the banks based in order listed in index

        :param list[Indexable] indexables: Banks that will be reordered based in index file
        :return list[Bank]: Banks reordered
        """
        try:
            data = Persistence.read(self.path, create_file=True)
        except ValueError:
            data = {}

        return self.load_data(data, indexables)

    def load_data(self, json_data, indexables):
        new_list = []
        indexables_hash = {uuid: bank for (uuid, bank) in map(lambda bank: (bank.uuid, bank), indexables)}

        for indexable in json_data:
            if indexable['uuid'] in indexables_hash:
                new_list.append(indexables_hash[indexable['uuid']])
                del indexables_hash[indexable['uuid']]

        return new_list + sorted(indexables_hash.values(), key=lambda indexable: indexable.simple_identifier)

    def save(self, indexables):
        data = self.generate_data(indexables)
        Persistence.save(self.path, data)

    def generate_data(self, indexables):
        data = []

        for index, indexable in enumerate(indexables):
            item = self.generate_index_data(index, indexable)
            data.append(item)

        return data

    def generate_index_data(self, index, indexable):
        item = {
            'index': index,
            'uuid': indexable.uuid,
            'identifier': indexable.simple_identifier
        }
        return item




# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from pluginsmanager.model.port import Port


class AudioPort(Port, metaclass=ABCMeta):
    """
    Port is a parent abstraction for audio inputs and audio outputs
    """

    @property
    def connection_class(self):
        """
        :return Connection: Class used for connections in this port
        """
        from pluginsmanager.model.connection import Connection
        return Connection



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid


class Indexable(object):
    def __init__(self):
        self._uuid = str(uuid.uuid4())

    @property
    def uuid(self):
        """
        :return string: Unique identifier used for saves
        """
        return self._uuid

    @property
    def simple_identifier(self):
        """
        Simple identifier for index file human comprehension and reordering
        :return string:
        """
        pass



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.system.system_effect import SystemEffect



class SystemEffectBuilder(object):
    """
    Automatic system physical ports detection

    :param JackClient jack_client: :class:`.JackClient` instance that will get the information to
                                   generate :class:`.SystemEffect`
    """
    def __init__(self, jack_client):
        self.client = jack_client

    def build(self):
        inputs  = (port.shortname for port in self.client.audio_inputs)
        outputs = (port.shortname for port in self.client.audio_outputs)
        midi_inputs  = (port.shortname for port in self.client.midi_inputs)
        midi_outputs = (port.shortname for port in self.client.midi_outputs)

        return SystemEffect('system', outputs, inputs, midi_outputs, midi_inputs)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import threading


class ManagerScopes(object):

    def __init__(self):
        self._scopes = {}

    @property
    def current(self):
        identifier = threading.current_thread().ident

        if identifier not in self._scopes:
            self._scopes[identifier] = Scope()

        return self._scopes[identifier]

    def enter(self, identifier):
        """
        Open a scope.

        Informs that changes occurs by the ``identifier`` and isn't necessary
        informs the changes for identifier

        :param identifier: Identifier for instance that causes changes
        """
        self.current.enter(identifier)

    def exit(self):
        """
        Closes the last scope added
        """
        self.current.exit()


class Scope(object):

    def __init__(self):
        self._scope = collections.deque()

    def enter(self, identifier):
        self._scope.append(identifier)

    def exit(self):
        self._scope.pop()

    @property
    def identifier(self):
        try:
            return self._scope[-1]
        except IndexError:
            return None



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from pluginsmanager.model.midi_port import MidiPort


class MidiInput(MidiPort, metaclass=ABCMeta):
    """
    MidiInput is the medium in which the midi input port will go into
    effect to be processed.

    For obtains the inputs::

        >>> cctonode
        <Lv2Effect object as 'CC2Note'  active at 0x7efe5480af28>
        >>> cctonode.midi_inputs
        (<Lv2MidiInput object as MIDI In at 0x7efe54535dd8>,)

        >>> midi_input = cctonode.midi_inputs[0]
        >>> midi_input
        <Lv2MidiInput object as MIDI In at 0x7efe54535dd8>

        >>> symbol = midi_input.symbol
        >>> symbol
        'midiin'

        >>> cctonode.midi_inputs[symbol] == midi_input
        True

    For connections between effects, see :meth:`~pluginsmanager.model.pedalboard.Pedalboard.connect()`
    and :meth:`~pluginsmanager.model.pedalboard.Pedalboard.disconnect()` :class:`.Pedalboard` class methods.

    :param Effect effect: Effect of midi input
    """

    @property
    def index(self):
        """
        :return: MidiInput index in the your effect
        """
        return self.effect.midi_inputs.index(self)



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from unittest.mock import MagicMock


class Port(metaclass=ABCMeta):
    """
    Port is a parent abstraction for inputs and outputs

    :param Effect effect: Effect that contains port
    """

    def __init__(self, effect):
        self._effect = effect
        self.observer = MagicMock()

    @property
    @abstractmethod
    def symbol(self):
        """
        :return: Identifier for this port
        """
        pass

    @property
    def effect(self):
        """
        :return: Effect that this port is related
        """
        return self._effect

    @property
    def json(self):
        """
        Get a json decodable representation

        :return dict: json representation
        """
        return self.__dict__

    @property
    @abstractmethod
    def index(self):
        """
        :return: Index in the effect related based in your category.
                 As example, if this port is a :class:`input`, the
                 index returns your position in the inputs ports.
        """
        pass

    @property
    def __dict__(self):
        return {
            'effect': self.effect.index,
            'symbol': self.symbol,
            'index': self.index,
        }

    def __repr__(self):
        return "<{} object as {} at 0x{:x}>".format(
            self.__class__.__name__,
            str(self),
            id(self)
        )

    @property
    @abstractmethod
    def connection_class(self):
        """
        :return: Class used for connections in this port
        """
        pass



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod


class AudioPortBuilder(metaclass=ABCMeta):
    """
    Extracts the inputs and outputs of an effect defined in a json.

    Use it to get the `AudioPorts`_ (inputs and outputs) to build a connection correctly.

    .. _AudioPorts: http://lv2plug.in/ns/lv2core/#AudioPort
    """

    @abstractmethod
    def build_input(self, json):
        """
        :return Input: Input of an effect defined in json
        """
        pass

    @abstractmethod
    def build_output(self, json):
        """
        :return Output: Output of an effect defined in json
        """
        pass

    @abstractmethod
    def build_midi_input(self, json):
        """
        :return MidiInput: MidiInput of an effect defined in json
        """
        pass

    @abstractmethod
    def build_midi_output(self, json):
        """
        :return MidiOutput: MidiOutput of an effect defined in json
        """
        pass


class EffectBuilder(metaclass=ABCMeta):

    @abstractmethod
    def build(self, json):
        pass



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.param import Param


class Lv2Param(Param):
    """
    Representation of a Lv2 `input control port`_ instance.

    For general input use, see :class:`.Param` class documentation.

    :param Lv2Effect effect: Effect that contains the param
    :param dict param: *input control port* json representation

    .. _input control port: http://lv2plug.in/ns/lv2core/#Parameter
    """

    def __init__(self, effect, param):
        super(Lv2Param, self).__init__(effect, param['ranges']['default'])
        self._param = param

    @property
    def maximum(self):
        return self._param['ranges']['maximum']

    @property
    def minimum(self):
        return self._param['ranges']['minimum']

    @property
    def symbol(self):
        return self._param['symbol']

    @property
    def __dict__(self):
        dictionary = super(Lv2Param, self).__dict__
        dictionary['index'] = self._param['index']

        return dictionary



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pluginsmanager.util.builder.builder import AudioPortBuilder, EffectBuilder


class Lv2AudioPortBuilder(AudioPortBuilder):
    """
    Extracts the :class:`.Lv2Input`s and :class:`.Lv2Outputs`s of an :class:`.Lv2Effect` defined in a json.
    """

    def __init__(self, pedalboard):
        self.pedalboard = pedalboard

    def build_input(self, json):
        symbol = json['symbol']
        return self.get_effect(json).inputs[symbol]

    def build_output(self, json):
        symbol = json['symbol']
        return self.get_effect(json).outputs[symbol]

    def build_midi_input(self, json):
        symbol = json['symbol']
        return self.get_effect(json).midi_inputs[symbol]

    def build_midi_output(self, json):
        symbol = json['symbol']
        return self.get_effect(json).midi_outputs[symbol]

    def get_effect(self, json):
        effect_index = json['effect']
        return self.pedalboard.effects[effect_index]


class Lv2EffectBuilder(EffectBuilder):
    def __init__(self, lv2_effect_builder):
        """
        :param model.lv2.lv2_effect_builder.Lv2EffectBuilder lv2_effect_builder:
        """
        self.builder = lv2_effect_builder

    def build(self, json):
        effect = self.builder.build(json['plugin'])

        for param, param_json in zip(effect.params, json['params']):
            param.value = param_json['value']

        effect.active = json['active']

        return effect



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Lv2Plugin(object):

    def __init__(self, json):
        self._json = json

    def __getitem__(self, key):
        """
        :param string key: Property key
        :return: Returns a Plugin property
        """
        return self.json[key]

    @property
    def json(self):
        """
        Json decodable representation of this plugin based in moddevices `lilvlib`_.

        .. _lilvlib: https://github.com/moddevices/lilvlib
        """
        return self._json

    def __str__(self):
        return self['name']

    def __repr__(self):
        return "<{} object as {} at 0x{:x}>".format(
            self.__class__.__name__,
            str(self),
            id(self)
        )

    @property
    def version(self):
        return '' if 'version' not in self.json else self.json['version']



# Copyright 2017 SrMouraSilva
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pluginsmanager.model.connection import Connection
from pluginsmanager.model.midi_port import MidiPort


class MidiConnection(Connection):
    """
    :class:`.MidiConnection` represents a connection between two
    distinct effects by your :class:`.MidiPort` (effect :class:`.MidiOutput` with effect :class:`.MidiInput`)::

    >>> californication = Pedalboard('Californication')
    >>> californication.append(driver)
    >>> californication.append(reverb)

    >>> output_port = cctonode1.midi_outputs[0]
    >>> input_port = cctonode2.midi_inputs[0]

    >>> californication.connections.append(MidiConnection(output_port, input_port))

    Another way to use implicitly connections::

    >>> californication.connect(output_port, input_port)

    :param MidiOutput output_port: MidiOutput port that will be connected with midi input port
    :param MidiInput input_port: MidiInput port that will be connected with midi output port
    """

    @property
    def ports_class(self):
        """
        :return class: Port class that this connection only accepts
        """
        return MidiPort

    @property
    def __dict__(self):
        dictionary = super(MidiConnection, self).__dict__
        dictionary['type'] = 'midi'

        return dictionary
